// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt

#ifndef __LUXCORE_OCTREE_H__
#define __LUXCORE_OCTREE_H__

//////////////////////////////////////////////////////////////////////////
// OcTree
//
// NOT THREADSAFE!! except traverse & collide with per-thread stacks

#include <luxplatform/luxplatform.h>

#ifdef __cplusplus
extern "C"{
#endif

#define OCTREE_MAX_DEPTH		31
#define OCTREE_MAX_STACKITEMS	((OCTREE_MAX_DEPTH+1)*8)

typedef struct OcCenterBox_s
{
	float		center[3];
	float		size[3];
	float		min[3];	// order must not change (minmax)
	float		max[3];
}OcCenterBox_t;

//typedef struct OcDistanceInfo_s {
//	float		distance;
//	void		*ptr;
//} OcDistanceInfo_t;

typedef struct OcContainerBox_s {
	OcCenterBox_t	box;
	void			*data;
	//OcDistanceInfo_t *contactData;
	//int				contactCount;
	//int				contactID;
} OcContainerBox_t;

typedef struct OcListNode_s {
	OcContainerBox_t	*container;
	struct OcNode_s		*owner;
	struct OcListNode_s *next, *prev;
} OcListNode_t;

typedef struct OcNode_s {
	OcCenterBox_t		volume;
	float				min[3];	// order must not change (minmax)
	float				max[3];
	struct OcNode_s		*parent;
	struct OcNode_s		*childs[8];
	OcListNode_t		*list;
	int					listCount;
} OcNode_t;

typedef struct OcTree_s*	OcTreePTR;


typedef struct OcTravStackItem_s {
	OcNode_t	*node;
	int			depth;
} OcTravStackItem_t;


typedef struct OcTravStack_s {
	OcTravStackItem_t	items[OCTREE_MAX_STACKITEMS];
}OcTravStack_t;


	// called after OcNode is created
	// can be used to write to appended customdata per node
	// custom data will start after sizeof(OcNode_t)
typedef void (OcNodeBuild_fn) (OcNode_t* node,void *upvalue);

	// container - containerbox of 'other' tree
	// node - the Ocnode it intersects with
	// depth - current depth in the tree
	// return 0 to stop branching
typedef int (OcContactTest_fn) (OcContainerBox_t* occontainer, OcNode_t* node, int depth,void *upvalue);


	// node - the Ocnode it intersects with
	// depth - how far we are in the tree
	// return 0 to stop branching
typedef int (OcTraverse_fn) (OcNode_t* node, int depth, void *upvalue);

	// count - number of list items in the current ocnode
	// list - listnodes in the current ocnode
	// return 0 to stop branching
typedef int (OcListNodeTraverse_fn) (int listcount, OcListNode_t* list, void *upvalue);

	// for drawing the octree
typedef void (OcDrawBox_fn) (const float minb[3], const float maxb[3], const float color[4]);

//////////////////////////////////////////////////////////////////////////
// OcNode

	// calls listcollector for each list in self or in one of its childs. Branching is 
	// stoped if the collector returns 0
void OcNode_getLists_recursive (OcNode_t	*self, OcListNodeTraverse_fn *listcollector, void *upvalue);


////////////////////////////////////////////////////////////////////////////////
// OcTree


	// constructor
OcTreePTR OcTree_new ();
	// destructor
void OcTree_free (OcTreePTR	self);

	// after adding ALL items to the tree, this function must be called
	// the allocated memory per OcNode can be expanded to store custom data
	// ocNodeBuild can be passed to be called after OcNode was allocated
void OcTree_build(OcTreePTR	self, int maxDepth, size_t nodeExtraMem, OcNodeBuild_fn* nodefunc, void *upvalue);

	// allows extra data to be copied at end of "add"
	// Can only be changed after "new" or "reset"
void OcTree_containerExtraMem(OcTreePTR	self, size_t containerExtraMem);

	// adds an object to the tree, w,h,d are the diameters, so for a sphere of radius r, 
	// w = 2*r, ...
	// You can copy extra data (if OcTree has containerExtraMem)
	// to the pointer returned. This pointer will only be valid
	// until the next add!
void* OcTree_add (OcTreePTR	self, void *data, float x, float y, float z, 
	float w, float h, float d);

	// clears all objects from the tree - should be called after build() before add()
void OcTree_reset(OcTreePTR	self);

	// draws the octree via OpenGL, if maxdepth is -1 we will use self's maxdepth
void OcTree_draw (OcTreePTR	self, int fromdepth, int maxdepth, OcDrawBox_fn *drawfunc);

	// traverses the octree's node which intersects the given volume. 
	// each node has a list of nodes with your data. You must traverse the list
	// in order to check your listnodes. You can stop traversion of the current
	// brunch (but not the traversion).
	// if threadstack is NULL, default is used (ie not threadsafe)
void OcTree_traverse(OcTreePTR	self, OcTravStack_t *threadstack, OcTraverse_fn *traversefn,
	float cx, float cy, float cz, float w, float h, float d, void *upvalue);
	
	// similiar to traverse, if the traversing function returns 0, the branching is
	// stopped. Quickaccepts or rejects can be made this way
	// if threadstack is NULL, default is used (ie not threadsafe)
void OcTree_collide (OcTreePTR	self, OcTravStack_t *threadstack, OcTreePTR	other, OcContactTest_fn *contactTester, void *upvalue);


	// similar to above but for the whole tree
void OcTree_getLists(OcTreePTR	self,OcListNodeTraverse_fn *listcollector, void *upvalue);


#ifdef __cplusplus
};  
#endif

#endif

