// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt
//
// <Author: Christoph Kubisch>


#ifndef __LUXCORE_REFSYS_H__
#define __LUXCORE_REFSYS_H__

#include <luxplatform/luxplatform.h>

#ifdef __cplusplus
extern "C"{
#endif

	// A system to manage reference counted objects.
	// Prevent cyclic refs with addWeak and releaseWeak
	// The object is kept alive as long as users exist or was invalidated by user.
	// The reference is kept alive as long as weak users exist
	// (object itself being one weak user)

	// NOT THREADSAFE, uses MEMGENERIC

	//////////////////////////////////////////////////////////////////////////

	typedef struct ObjRefSys_s*  ObjRefSysPTR;
	typedef	struct ObjRef_s*	ObjRefPTR;
	typedef	void (ObjRefDeleteFn)(ObjRefPTR r);
	typedef	booln (ObjRefCheckDeleteFn)(ObjRefPTR r);

	enum ObjRefType_e{
		OBJREF_TYPE_DELETED = -1,
		OBJREF_TYPE_FREEALLOC = 0,
		OBJREF_TYPE_USERSTART,
	};
	typedef int32	ObjRefType_t;

	typedef struct ObjId_s{
		void			*ptr;
		ObjRefType_t	type;
	}ObjId_t;

	typedef struct ObjTypeInfo_s{
		ObjRefDeleteFn*	fnDelete;
#ifdef LUX_DEVBUILD
		const char*		name;
#endif
	}ObjTypeInfo_t;


	//////////////////////////////////////////////////////////////////////////
	// ObjRefSys

		// allocs mem & runs init
	LUX_API ObjRefSysPTR	ObjRefSys_new();

		// runs deinit &  deallocs mem
	LUX_API void			ObjRefSys_delete(ObjRefSysPTR sys);

		// standalone init/deinit
	LUX_API void			ObjRefSys_init(ObjRefSysPTR sys);

		// optionally calls func prior all delete operations
	LUX_API void			ObjRefSys_deinit(ObjRefSysPTR sys, ObjRefCheckDeleteFn* func);

	LUX_API int32			ObjRefSys_refCount(ObjRefSysPTR sys);

		// within sections delete by release is prevented
		// all deletes are postponed to leaving section
	LUX_API void ObjRefSys_pushNoDelete(ObjRefSysPTR sys);
	LUX_API void ObjRefSys_popNoDelete(ObjRefSysPTR sys);

		// register a new type
	ObjTypeInfo_t ObjTypeInfo_new(ObjRefDeleteFn* fnDelete, const char*	name);
	LUX_API void ObjRefSys_register(ObjRefSysPTR sys, ObjRefType_t type, struct ObjTypeInfo_s info);

		// usecounter and weakcounter inited at 1
	LUX_API ObjRefPTR	ObjRefSys_newRef(ObjRefSysPTR sys, ObjRefType_t type, void *ptr);
	void		ObjRefSys_releaseWeakRef(ObjRefSysPTR sys, ObjRefPTR cref);

		// returns true if ref remains valid
	booln		ObjRefSys_releaseUserRef(ObjRefSysPTR sys, ObjRefPTR cref);

		// sets internal state to deleted, clears ptr
		// leaves counters unchanged, calldestructor ignores NoDelete status!
		// returns true if there were non-weak links left, which means
		// it's quite unsafe to do!
	LUX_API booln		ObjRefSys_invalidateRef(ObjRefSysPTR sys,ObjRefPTR cref, booln calldestructor);
	

	//////////////////////////////////////////////////////////////////////////
	// ObjRef

		// returns ptr only if reference is still valid (wasn't deleted)
	ObjId_t*	ObjRef_getId(ObjRefPTR cref);

		// returns ptr (is NULL when already deleted, ie weak refs must check this)
	void*		ObjRef_getPtr(ObjRefPTR cref);

		// if you allow creation of ObjRefs with NULL ptrs, than you must
		// use this instead of getPtr
	booln		ObjRef_getSafe(ObjRefPTR cref, void **ptr);
		
		// increments weak counter
	void		ObjRef_addWeak(ObjRefPTR cref);

		// returns true on success (ref is valid)
	booln		ObjRef_addUser(ObjRefPTR cref);

		// set usecounter to 0 (only allowed when usecounter is 1)
		// useful if you want to return "newobjects", 
		// use with caution, will not call destructor nor change weakcounter.
	booln		ObjRef_makeVolatile(ObjRefPTR cref);

	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////
	// Inline & Details (usually not needed, don't mess with content!)

	typedef struct ObjRef_s{
		ObjId_t			id;
		int32			usecounter;
		int32			weakcounter;
	}ObjRef_t;


	// for refsys (only in case you want to create it statically)
#include <luxcore/contvector.h>
	typedef struct ObjRefAllocSys_s{
		ObjRef_t			*freerefs;
		struct ObjRefPage_s	*pagelist;
		size_t				lastinpage;
	}ObjRefAllocSys_t;


	typedef struct ObjRefSys_s{
		int32				nodelete;
		int32				count;
		ObjRefAllocSys_t	allocator;
		ContVector_t		typeinfovec;
		ObjTypeInfo_t		*typeinfos;
		ContVector_t		nodelvec;
	}ObjRefSys_t;

	LUX_API void		ObjRefSys_deleteRef(ObjRefSysPTR sys, ObjRefPTR cref);
	LUX_API void		ObjRefSys_deleteAlloc(ObjRefSysPTR sys, ObjRefPTR cref);

	LUX_INLINE int32 ObjRefSys_refCount(ObjRefSysPTR sys){
		return sys->count;
	}

	LUX_INLINE ObjTypeInfo_t ObjTypeInfo_new(ObjRefDeleteFn* fnDelete, const char* name)
	{
		ObjTypeInfo_t	info;
		info.fnDelete = fnDelete;
#ifdef LUX_DEVBUILD
		info.name = name;
#endif
		return info;
	}

	LUX_INLINE void	ObjRefSys_releaseWeakRef(ObjRefSysPTR sys, ObjRefPTR cref)
	{
		if (--cref->weakcounter == 0 ){
			ObjRefSys_deleteAlloc(sys,cref);
		}
	}

	LUX_INLINE booln	ObjRefSys_releaseUserRef(ObjRefSysPTR sys,ObjRefPTR cref)
	{
		booln valid = (cref->id.type >= OBJREF_TYPE_USERSTART);
		if( --cref->usecounter == 0 )
		{
			ObjRefSys_deleteRef(sys,cref);
			valid = FALSE;
		}

		return valid;
	}

	LUX_INLINE booln	ObjRef_makeVolatile(ObjRefPTR cref)
	{
		if (cref->id.type < OBJREF_TYPE_USERSTART || cref->usecounter != 1) return FALSE;
		cref->usecounter = 0;
		return TRUE;
	}
	LUX_INLINE booln	ObjRef_getSafe(ObjRefPTR cref, void **ptr)
	{
		booln valid = (cref && cref->id.type >= OBJREF_TYPE_USERSTART);
		*ptr = valid ? cref->id.ptr : NULL;
		return valid;
	}

	LUX_INLINE void*	ObjRef_getPtr(ObjRefPTR cref)
	{
		return cref->id.ptr;
	}
	LUX_INLINE ObjId_t*	ObjRef_getId(ObjRefPTR cref)
	{
		return cref->id.type >= OBJREF_TYPE_USERSTART ? (ObjId_t*)cref : NULL;
	}

	LUX_INLINE void	ObjRef_addWeak(ObjRefPTR cref)
	{
		cref->weakcounter++;
	}


	LUX_INLINE booln	ObjRef_addUser(ObjRefPTR cref)
	{
		if (cref->id.type < OBJREF_TYPE_USERSTART) return FALSE;

		cref->usecounter++;
		return TRUE;
	}


#ifdef __cplusplus
}
#endif

#endif
