// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt

#include <luxcore/contoctree.h>
#include <luxcore/memorygeneric.h>
#include <luxcore/contmacrolinkedlist.h>
#include <luxcore/debug.h>
#include <luxmath/vector4.h>



//////////////////////////////////////////////////////////////////////////
// DEFS

#define OCTREE_NODELIST_MIN			4

typedef struct OcTree_s {
	OcNode_t 			*root;
	float				min[3];
	float				max[3];
	int					maxdepth;
	int					memsize;
	int					memuse;
	int					containerpos;
	byte				*memblock;

	// special iterators/data 
	OcContactTest_fn	*contactTester;
	OcNodeBuild_fn		*nodeBuild;
	size_t				nodeExtraAlloc;
	size_t				containerExtraAlloc;
	void*				upvalue;
} OcTree_t;

static OcTravStack_t l_traverseStack;

static OcNode_t* OcTree_getNode(OcTree_t* self, float x, float y, float z,
	float w, float h, float d);
static OcListNode_t* OcTree_getListNode (OcTree_t *self, OcContainerBox_t *box);



////////////////////////////////////////////////////////////////////////////////
// OcCenterBox

LUX_INLINE void OcCenterBox_init (OcCenterBox_t *self, float x, float y, float z, float w, float h, float d)
{
	self->center[0] = x;
	self->center[1] = y;
	self->center[2] = z;
	self->size[0] = w;
	self->size[1] = h;
	self->size[2] = d;

	self->min[0] = x-(w*0.5f);
	self->min[1] = y-(h*0.5f);
	self->min[2] = z-(d*0.5f);
	self->max[0] = x+(w*0.5f);
	self->max[1] = y+(h*0.5f);
	self->max[2] = z+(d*0.5f);
}


LUX_INLINE booln OcCenterBox_intersects(OcCenterBox_t *self, OcCenterBox_t *other){
	  return (	(self)->min[0]<=(other)->max[0] && 
				(self)->min[1]<=(other)->max[1] &&
				(self)->min[2]<=(other)->max[2] &&
				(self)->max[0]>=(other)->min[0] &&
				(self)->max[1]>=(other)->min[1] &&
				(self)->max[2]>=(other)->min[2]	);
}

LUX_INLINE booln OcNode_intersectsCenter(OcNode_t *self, OcCenterBox_t *other){
	return (	(self)->min[0]<=(other)->max[0] && 
				(self)->min[1]<=(other)->max[1] &&
				(self)->min[2]<=(other)->max[2] &&
				(self)->max[0]>=(other)->min[0] &&
				(self)->max[1]>=(other)->min[1] &&
				(self)->max[2]>=(other)->min[2]	);
}



////////////////////////////////////////////////////////////////////////////////
// OcNode

static OcNode_t* OcNode_add(OcNode_t *self, OcListNode_t *node)
{
	//static float t;
	if (node == NULL) return NULL;

	ListNode_addLast(self->list,node);
	//t = node->container->box.center[0]-node->container->box.size[0]/2;
	self->min[0] = MIN(self->min[0],node->container->box.min[0]);
	//t+= node->container->box.size[0];
	self->max[0] = MAX(self->max[0],node->container->box.max[0]);

	//t = node->container->box.center[1]-node->container->box.size[1]/2;
	self->min[1] = MIN(self->min[1],node->container->box.min[1]);
	//t+= node->container->box.size[1];
	self->max[1] = MAX(self->max[1],node->container->box.max[1]);

//	t = node->container->box.center[2]-node->container->box.size[2]/2;
	self->min[2] = MIN(self->min[2],node->container->box.min[2]);
//	t+= node->container->box.size[2];
	self->max[2] = MAX(self->max[2],node->container->box.max[2]);

	node->owner = self;
	self->listCount++;

	return self;
}

static void OcNode_build(OcNode_t *self, OcTree_t *tree, int restDepth)
{

	int pos,depth;

	pos = 1;
	l_traverseStack.items[0].node = self;
	l_traverseStack.items[0].depth = restDepth;

	while (pos>0) {
		depth = l_traverseStack.items[--pos].depth;
		self  = l_traverseStack.items[pos].node;

		if (self->listCount>OCTREE_NODELIST_MIN)
		{
			OcListNode_t *browse,*removed;
			int sector,a,b,c, cont;
			float cx,cy,cz,w,h,d,max;

			if (self->list == NULL) {
				return;
			}

			cx = (self->max[0]+self->min[0])*0.5f;
			cy = (self->max[1]+self->min[1])*0.5f;
			cz = (self->max[2]+self->min[2])*0.5f;
			w = self->max[0]-self->min[0];
			h = self->max[1]-self->min[1];
			d = self->max[2]-self->min[2];

			
			#define MAX3(a,b,c) (((a)>(b))?((a)>(c))?(a):(c):((b)>(c))?(b):(c))
			max = MAX3(w,h,d);

			// the bounding box seems to be planar => make it quadtreelike by modifying the bounds
			if (max>2*w) {
				w*=2.01f;
				cx+=w*0.25f;
			}
			if (max>2*h) {
				h*=2.01f;
				cy+=h*0.25f;
			}
			if (max>2*d) {
				d*=2.01f;
				cz+=d*0.25f;
			}

			self->volume.min[0] = self->max[0];	// set volume to invalid values
			self->volume.min[1] = self->max[1];
			self->volume.min[2] = self->max[2];
			self->volume.max[0] = self->min[0];
			self->volume.max[1] = self->min[1];
			self->volume.max[2] = self->min[2];

			browse = self->list;
			do {
			//	luxprintf("%i %p %p\n",restDepth,browse,self->list);
				cont = 0;
				if (browse->container->box.size[0]*2.0f<w &&		// cb removed || 
					browse->container->box.size[1]*2.0f<h &&
					browse->container->box.size[2]*2.0f<d)
				{
					a = cx > browse->container->box.center[0];
					b = cy > browse->container->box.center[1];
					c = cz > browse->container->box.center[2];
					sector = (a<<2)|(b<<1)|c;
					//if (sector)
					{
						removed = browse;
						browse = browse->prev;
						if (removed == self->list) {
							self->list = self->list->prev;
						}
						removed->next->prev = removed->prev;
						removed->prev->next = removed->next;
						removed->next = removed->prev = removed;
						if (self->childs[sector] == NULL) {
							OcNode_t *newnode = OcTree_getNode(tree,
								cx-w*((float)a-0.5f)*0.5f,
								cy-h*((float)b-0.5f)*0.5f,
								cz-d*((float)c-0.5f)*0.5f,
								w*0.5f,
								h*0.5f,
								d*0.5f);

							LUX_ASSERT(newnode);

							newnode->parent = self;
							self->childs[sector] = newnode;
						}
						//if (self->childs[sector] == NULL)
						//	break;
						OcNode_add(self->childs[sector],removed);
						
						self->listCount--;
						if (removed == browse) {
							self->list = NULL;
							break;
						}
						cont = 1;
						continue;
					}
				}

				//browse->owner = self;
				self->volume.min[0] = MIN(self->volume.min[0],browse->container->box.min[0]);
				self->volume.min[1] = MIN(self->volume.min[1],browse->container->box.min[1]);
				self->volume.min[2] = MIN(self->volume.min[2],browse->container->box.min[2]);
				self->volume.max[0] = MAX(self->volume.max[0],browse->container->box.max[0]);
				self->volume.max[1] = MAX(self->volume.max[1],browse->container->box.max[1]);
				self->volume.max[2] = MAX(self->volume.max[2],browse->container->box.max[2]);
				browse = browse->next;
			} while (self->list != browse || cont);

		}
		else{
			// cb,added, otherwise self->volume is incorrect
			OcListNode_t *browse = self->list;

			self->volume.min[0] = self->max[0];	// set volume to invalid values
			self->volume.min[1] = self->max[1];
			self->volume.min[2] = self->max[2];
			self->volume.max[0] = self->min[0];
			self->volume.max[1] = self->min[1];
			self->volume.max[2] = self->min[2];

			
			do{
				self->volume.min[0] = MIN(self->volume.min[0],browse->container->box.min[0]);
				self->volume.min[1] = MIN(self->volume.min[1],browse->container->box.min[1]);
				self->volume.min[2] = MIN(self->volume.min[2],browse->container->box.min[2]);
				self->volume.max[0] = MAX(self->volume.max[0],browse->container->box.max[0]);
				self->volume.max[1] = MAX(self->volume.max[1],browse->container->box.max[1]);
				self->volume.max[2] = MAX(self->volume.max[2],browse->container->box.max[2]);
				browse = browse->next;
			} while (self->list != browse);
			
		}
#if 0
		LUX_ASSERT(self->min[0] <= self->volume.min[0]);
		LUX_ASSERT(self->min[1] <= self->volume.min[1]);
		LUX_ASSERT(self->min[2] <= self->volume.min[2]);

		LUX_ASSERT(self->max[0] >= self->volume.max[0]);
		LUX_ASSERT(self->max[1] >= self->volume.max[1]);
		LUX_ASSERT(self->max[2] >= self->volume.max[2]);
		{
			OcListNode_t *browse = self->list;
			if (browse)
				do{	
					LUX_ASSERT(browse->container->box.min[0] >= self->volume.min[0]);
					LUX_ASSERT(browse->container->box.min[1] >= self->volume.min[1]);
					LUX_ASSERT(browse->container->box.min[2] >= self->volume.min[2]);

					LUX_ASSERT(browse->container->box.max[0] <= self->volume.max[0]);
					LUX_ASSERT(browse->container->box.max[1] <= self->volume.max[1]);
					LUX_ASSERT(browse->container->box.max[2] <= self->volume.max[2]);

					browse = browse->next;
				} while (self->list != browse);
		}
#endif

		if (depth>0) {
			int i;
			for (i = 0;i < 8;i++)
			{
				if (self->childs[i]!=NULL) {
					l_traverseStack.items[pos].node    = self->childs[i];
					l_traverseStack.items[pos++].depth = depth - 1;
					LUX_ASSERT(OCTREE_MAX_STACKITEMS!=pos);// stack overflow of local stack
				}
			}
		}
	}
}

static void OcNode_traverse(OcNode_t *self, OcTravStack_t *threadstack, OcTraverse_fn *traversefn,
	int depth,OcCenterBox_t *box, void *upvalue)
{
	int i;
	int pos;

	pos = 1;
	threadstack->items[0].node = self;
	threadstack->items[0].depth = 0;

	while (pos>0) {
		depth = threadstack->items[--pos].depth;
		self  = threadstack->items[pos].node;
		if (OcCenterBox_intersects(&self->volume,box) &&
			!traversefn(self,depth,upvalue)) continue;

		for (i=0;i<8;i++)
			if (self->childs[i]!=NULL) {
				threadstack->items[pos].node = self->childs[i];
				threadstack->items[pos++].depth = depth + 1;
				LUX_ASSERT(OCTREE_MAX_STACKITEMS!=pos);// stack overflow of local stack
			}
	}
}

static void OcNode_addContact (OcNode_t *self, OcTravStack_t *threadstack, OcContainerBox_t *contbox,
	OcContactTest_fn *tester, int depth, void *upvalue)
{
	int i;
	int pos;

	pos = 1;
	threadstack->items[0].node = self;
	threadstack->items[0].depth = 0;

	while (pos>0) {
		depth = threadstack->items[--pos].depth;
		self = threadstack->items[pos].node;

		if (!tester(contbox,self,depth,upvalue)) continue;

		for (i = 0;i<8;i++)
			if (self->childs[i]!=NULL)
				if (OcNode_intersectsCenter(self->childs[i],&contbox->box))
			{
				threadstack->items[pos].node = self->childs[i];
				threadstack->items[pos++].depth = depth + 1;
				LUX_ASSERT(OCTREE_MAX_STACKITEMS!=pos); // stack overflow of local stack
			}
	}
}

static void OcNode_draw(OcNode_t *self,int depth,OcDrawBox_fn *drawfunc, const int drawfrom, const int drawto)
{
	Vector4 color;
	int sector;
	OcListNode_t *oclist;

	if (depth < drawfrom || depth > drawto);
	else{
		float range = (float)(drawto-drawfrom);
		float depthcolor = (1-((depth-drawfrom)/((range<1)? 1 : range))*0.8f);
		
		Vector4Set(color,depthcolor,0.0f,0.0f,1.0f);
		drawfunc(self->volume.min,self->volume.max,color);
		if (self->list!=NULL) {
			Vector4Set(color,0.0f,depthcolor,0.0f,1.0f);
			drawfunc(self->min,self->max,color);

			ListNode_forEach(self->list,oclist)
				Vector4Set(color,0.0f,0.0f,depthcolor,1.0f);
				drawfunc(oclist->container->box.min,oclist->container->box.max,color);
			ListNode_forEachEnd(self->list,oclist);
		}
	}
	for (sector = 0;sector < 8;sector++)
		if (self->childs[sector]!=NULL)
			OcNode_draw(self->childs[sector],depth+1,drawfunc,drawfrom,drawto);
}

static OcNode_t* OcNode_init(OcNode_t *self,
	float x,float y,float z, float w, float h, float d)
{
	int i;

	if (self==NULL) return NULL;
	OcCenterBox_init(&self->volume,x,y,z,w,h,d);
	self->min[0] = self->volume.min[0];
	self->min[1] = self->volume.min[1];
	self->min[2] = self->volume.min[2];
	self->max[0] = self->volume.max[0];
	self->max[1] = self->volume.max[1];
	self->max[2] = self->volume.max[2];
	self->listCount = 0;
	for (i=0;i<8;i++) self->childs[i] = NULL;
	self->list = NULL;

	return self;
}

////////////////////////////////////////////////////////////////////////////////
// OcListNode functions
static OcListNode_t* OcListNode_init (OcListNode_t *self, OcContainerBox_t *container)
{
	if (self==NULL) return NULL;
	self->container = container;
	ListNode_init(self);

	return self;
}

////////////////////////////////////////////////////////////////////////////////
// Container functions
static OcContainerBox_t* OcContainerBox_init (OcContainerBox_t* self, void *data,
	float x, float y, float z, float w, float h, float d)
{
	OcCenterBox_init(&self->box,x,y,z,w,h,d);
	self->data = data;

	return self;
}

////////////////////////////////////////////////////////////////////////////////
// OcTree functions

void OcTree_reset(OcTree_t *self)
{
	self->memuse = 0;
	self->containerpos = 0;
	self->root = NULL;
	self->containerExtraAlloc = 0;
	self->nodeExtraAlloc = 0;
}

static void* OcTree_malloc(OcTree_t* self,int size)
{
	byte* pos;

	if (self->memuse+size >= self->memsize) {
		int oldsize = self->memsize;
		self->memsize=self->memsize*2+size;
		self->memblock = (byte*)genrealloc(self->memblock,self->memsize,oldsize);
//		luxprintf("Dynamic resizing in add => %i\n",self->memsize);
		if (self->containerpos!=0) return NULL;
	}
	pos = &self->memblock[self->memuse];
	self->memuse+=size;

	return pos;
}

static OcNode_t* OcTree_getNode(OcTree_t* self, float x, float y, float z,
	float w, float h, float d)
{
	OcNode_t* node = OcNode_init((OcNode_t*)OcTree_malloc(self,sizeof(OcNode_t)+self->nodeExtraAlloc),x,y,z,w,h,d);
	if (self->nodeBuild) self->nodeBuild( node, self->upvalue);

	return node;
}

static OcListNode_t* OcTree_getListNode (OcTree_t *self, OcContainerBox_t *box)
{
	return OcListNode_init((OcListNode_t*)OcTree_malloc(self,sizeof(OcListNode_t)),box);
}

void OcTree_traverse(OcTree_t *self, OcTravStack_t *threadstack, OcTraverse_fn *traversefn,
	float cx, float cy, float cz, float w, float h, float d, void *upvalue)
{
	OcCenterBox_t box;

	if (self->root == NULL) return;
	box.center[0] = cx;
	box.center[1] = cy;
	box.center[2] = cz;
	box.size[0] = w;
	box.size[1] = h;
	box.size[2] = d;
	box.min[0] = cx-w*0.5f;
	box.min[1] = cy-h*0.5f;
	box.min[2] = cz-d*0.5f;
	box.max[0] = cx+w*0.5f;
	box.max[1] = cy+h*0.5f;
	box.max[2] = cz+d*0.5f;

	threadstack = threadstack ? threadstack : &l_traverseStack;
	OcNode_traverse(self->root,threadstack,traversefn,0,&box,upvalue);
}

void OcTree_build(OcTree_t *self, int maxDepth, size_t nodeExtraMem, OcNodeBuild_fn* nodefunc, void *upvalue)
{
	float cx,cy,cz,w,h,d;
	byte *box;
	byte *fin;

	size_t occontsize = sizeof(OcContainerBox_t) + self->containerExtraAlloc;

	if (self->memuse == 0) 
		return;

	self->maxdepth = MIN(OCTREE_MAX_DEPTH,maxDepth);
	cx = (self->min[0]+self->max[0])*0.5f;
	cy = (self->min[1]+self->max[1])*0.5f;
	cz = (self->min[2]+self->max[2])*0.5f;
	w = self->max[0]-self->min[0];
	h = self->max[1]-self->min[1];
	d = self->max[2]-self->min[2];

	if (self->memsize/self->memuse<4) {
		int oldsize = self->memsize;
		self->memsize*=4;
		self->memblock = (byte*)genrealloc(self->memblock,self->memsize,oldsize);
	}

	self->nodeExtraAlloc = nodeExtraMem;
	self->nodeBuild = nodefunc;
	self->upvalue = upvalue;

	self->containerpos = self->memuse;
	box = self->memblock;
	fin = self->memblock + self->containerpos;

	self->root = OcTree_getNode(self,cx,cy,cz,w,h,d);
	self->root->parent = NULL;
	while (box < fin) {
		OcNode_add(self->root,OcTree_getListNode(self,(OcContainerBox_t*)box));
		box += occontsize;
	}
	if (maxDepth>1) 
		OcNode_build(self->root,self,maxDepth-1);
}

void OcTree_collide (OcTree_t *self, OcTravStack_t *threadstack, OcTree_t *other, OcContactTest_fn *contactTester,void *upvalue)
{
	byte *box = other->memblock;
	byte *fin = other->memblock + other->containerpos;
	size_t occontsize = sizeof(OcContainerBox_t) + other->containerExtraAlloc;

	if (!contactTester || !self->root)
		return;

	threadstack = threadstack ? threadstack : &l_traverseStack;
	while (box < fin) {
		OcNode_addContact(self->root,threadstack,(OcContainerBox_t*)box,contactTester,0,upvalue);
		box += occontsize;
	}
}

void OcNode_getLists_recursive (OcNode_t *self, OcListNodeTraverse_fn *listcollector, void *upvalue)
{
	int i;
	listcollector(self->listCount, self->list, upvalue);
	for (i=0;i<8;i++)
		if (self->childs[i]!=NULL)
			OcNode_getLists_recursive(self->childs[i],listcollector,upvalue);
}

void OcTree_getLists(OcTree_t *self,OcListNodeTraverse_fn *listcollector, void *upvalue)
{
	if (self->root){
		OcNode_getLists_recursive(self->root,listcollector,upvalue);
	}
}

void OcTree_containerExtraMem(OcTreePTR	self, size_t containerExtraMem)
{
	LUX_ASSERT(self->memuse == 0);
	self->containerExtraAlloc = containerExtraMem;		
}

void* OcTree_add (OcTree_t *self, void *data, float x, float y, float z,
	float w, float h, float d)
{
	OcContainerBox_t *container;

	// Octree_add not allowed after build
	LUX_ASSERT(self->containerpos == 0);

	if (self->memuse == 0)
	{	// first element that was added => set boundingbox to this object
		self->min[0] = x-w/2.0f;
		self->min[1] = y-h/2.0f;
		self->min[2] = z-d/2.0f;
		self->max[0] = x+w/2.0f;
		self->max[1] = y+h/2.0f;
		self->max[2] = z+d/2.0f;
	} else {
		float t;
		t = x-w/2.0f;	self->min[0] = MIN(self->min[0],t);
		t+= w;			self->max[0] = MAX(self->max[0],t);
		t = y-h/2.0f;	self->min[1] = MIN(self->min[1],t);
		t+= h;			self->max[1] = MAX(self->max[1],t);
		t = z-d/2.0f;	self->min[2] = MIN(self->min[2],t);
		t+= d;			self->max[2] = MAX(self->max[2],t);
	}

	container = OcContainerBox_init((OcContainerBox_t*)OcTree_malloc(self,
		sizeof(OcContainerBox_t) + self->containerExtraAlloc),data,x,y,z,w,h,d);

	return self->containerExtraAlloc ? container+1 : NULL;
}

OcTree_t* OcTree_new ()
{
	OcTree_t *self = NULL;

	self = (OcTree_t*)genzalloc(sizeof(OcTree_t));
	self->memsize = 1<<9;
	self->memuse = 0;
	self->containerpos = 0;
	self->memblock = (byte*)genzalloc(self->memsize);
	OcTree_reset(self);

	return self;
}

void OcTree_free (OcTree_t *self)
{
	genfree(self->memblock,self->memsize);
	genfree(self,sizeof(OcTree_t));
}

void OcTree_draw (OcTree_t *self, int fromdepth, int maxdepth, OcDrawBox_fn *drawfunc)
{
	Vector4 color = {0.0f,1.0f,0.0f,1.0f};
	if (maxdepth < 0)
		maxdepth = self->maxdepth;

	drawfunc(self->min,self->max,color);
	if (self->root!=NULL)
		OcNode_draw(self->root,1,drawfunc,fromdepth,maxdepth);
}