// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt

#include <luxcore/contstringmap.h>
#include <luxcore/contmacrolinkedlist.h>
#include <luxcore/memorygeneric.h>

#include <string.h>


//////////////////////////////////////////////////////////////////////////
// DEFS

typedef struct CharStrMap_s {
	void *dict;
	size_t maxNameLen;
	uint entryCount;
	char *nameBuffer;
} CharStrMap_t;

typedef struct KeyNode_s {
	int key;	// if key == 0, data == userdata, else data == next KeyNode

	union {
		void *data;
		struct KeyNode_s *child;
	};

	struct KeyNode_s LISTNODEVARS;
} KeyNode_t;


////////////////////////////////////////////////////////////////////////////////
// KeyNode

static KeyNode_t* KeyNode_new (const char *key, void *data)
{
	KeyNode_t *self;

	self = (KeyNode_t*) genzalloc(sizeof(KeyNode_t));
	self->key = *key;
	if (self->key == 0) self->data = data;
	else self->data = KeyNode_new(key+1,data);
	ListNode_init(self);

	return self;
}

static booln KeyNode_set (KeyNode_t *self, const char *key, void *data)
{
	KeyNode_t *browse;

START:
	browse = self;
	do {
		if (browse->key == *key)
		{
			if (*key != 0) {
				self = (KeyNode_t*)browse->data;
				key++;
				goto START;
			}
			else browse->data = data;
			return 0;
		}
		browse = ListNode_next(browse);
	} while (browse!=self);

	// node does not exist
	browse = KeyNode_new(key,data);
	ListNode_pushFront(self,browse);
	return 1;
}

static KeyNode_t* KeyNode_getKey (KeyNode_t *self, const char *key)
{
	KeyNode_t *browse;

KeyNode_getStart: // my jumppoint to restart loop - instead of recursion (4x faster)
	browse = self;
	do {
		if (browse->key == *key)
		{
			if (*key != 0) 	{
				self = (KeyNode_t*) browse->data;
				key++;
				goto KeyNode_getStart; // h0w 3v11 - but the quick & simple
			}
			else return browse;
		}
		browse = ListNode_next(browse);
	} while (browse!=self);

	return NULL;
}

static void* KeyNode_get (KeyNode_t *self, const char *key)
{	// returns the data associated with the given /key/
	KeyNode_t *found = KeyNode_getKey(self,key);
	if (found == NULL) return NULL;
	else return found->data;
}

static booln KeyNode_isset(KeyNode_t *self, const char *key)
{
	return KeyNode_getKey(self,key)!=NULL;
}

static void KeyNode_free(KeyNode_t *self)
{
	KeyNode_t *browse = self;

	do {
		if (browse->key != 0) KeyNode_free(browse->child);
		browse = ListNode_next(browse);
	} while (browse != self);

	while (self != NULL) {
		ListNode_popFront(self,browse);
		genfree(browse,sizeof(KeyNode_t));
	}
}

static KeyNode_t* KeyNode_unset(KeyNode_t *self, const char *key, booln *foundout)
{
	KeyNode_t *browse,
		*parent,	// the parent of branching
		*prevParent,// the parent of browse
		*child,		// next browsed element
		*min;		// the minor bound element that can be freed (will free all childs)
	booln found;

	found = FALSE;
	parent = NULL;
	prevParent = NULL;
	min = child = self;

KeyNode_getStart:
	browse = child;
	do
	{	// the loop that iterates over the listelements of same letterposition
		if (browse->key == *key)
		{
			found = TRUE;
			if (ListNode_next(browse) != browse)
			{	// if it's not the only element => I am branching here,
				// parent must not be freed
				// if we start on rootlayer we have no parent
				parent = prevParent ? prevParent : (KeyNode_t*)-1;
				min = browse;
			}
			if (*key != 0)
			{	// no termination
				child = browse->child;	// aquire new child
				prevParent = browse;	// update parent
				key++;
				goto KeyNode_getStart;	// continue with listiteration on next key-layer
			}
			else break;
		}
		browse = ListNode_next(browse);
	} while (browse!=child);

	*foundout = found;

	if (found){
		if (parent == NULL)
		{
			KeyNode_free(self);
			return NULL;
		}
		// rootlayer special case
		// we have no parent
		// it could be that we were the first, then next one is listhead
		if (parent == (void*)-1){
			self = (min == self) ? self->next : self;
			ListNode_remove(self,min);
		}
		else
			ListNode_remove(parent->child,min);
		KeyNode_free(min);
	}
	return self;
}

static void KeyNode_iterate (KeyNode_t *self, char *name, char *pos,
							 void* (*fn)(void*,const char *,void*), void *data)
{
	KeyNode_t *browse;

	ListNode_forEach(self,browse) {
		*pos = browse->key;
		if (*pos == 0)
		{
			data = fn(data,name, browse->data);
		} else {
			KeyNode_iterate((KeyNode_t*) browse->data, name,pos+1,fn,data);
		}
	} ListNode_forEachEnd(self,browse);
}


//////////////////////////////////////////////////////////////////////////
// CharStrMap

#ifdef MEMORY_GENERIC_STATS
CharStrMap_t*	CharStrMap_newTrace(const char *file, int line)
{
	CharStrMap_t *self = (CharStrMap_t*) MemoryGeneric_zallocStats(sizeof(CharStrMap_t),file,line);
#else
CharStrMap_t*	CharStrMap_new()
{
	CharStrMap_t *self = (CharStrMap_t*) genzalloc(sizeof(CharStrMap_t));
#endif

	self->dict = NULL;
	self->maxNameLen = 32;
	self->nameBuffer = (char*)genzalloc(self->maxNameLen);
	self->entryCount = 0;
	return self;
}

booln CharStrMap_set(CharStrMap_t *self, const char *key, void *value)
{
	int oldcount = self->entryCount;
	if (strlen(key)+1 > self->maxNameLen)
	{
		size_t oldsize = self->maxNameLen;
		self->maxNameLen = strlen(key)+1;
		self->nameBuffer = (char*)genrealloc(self->nameBuffer,self->maxNameLen,oldsize);
	}

	if (self->dict == NULL)
	{
		self->dict = KeyNode_new(key,value);
		self->entryCount = 1;
	}
	else self->entryCount += KeyNode_set((KeyNode_t*)self->dict,key,value);

	return self->entryCount-oldcount;
}
void*	CharStrMap_get(CharStrMap_t *self, const char *key)
{
	if (self->dict == NULL) return NULL;
	return KeyNode_get((KeyNode_t*)self->dict,key);
}

booln 	CharStrMap_unset(CharStrMap_t *self, const char *key)
{
	booln found = FALSE;

	if (self->dict)
		self->dict = KeyNode_unset((KeyNode_t*)self->dict, key,&found);

	if (found)
		self->entryCount--;

	return found;
}

booln 	CharStrMap_isset(CharStrMap_t *self, const char *key)
{
	if (self->dict==NULL) return 0;
	return KeyNode_isset((KeyNode_t*)self->dict,key);
}

static void*	CharStrMap_freeIter(void *data,const char *name, void* browse)
{
	void (*fnFree)(void*);
	if (data == NULL) return data;
	fnFree = (void (*)(void*))data;
	fnFree(browse);
	return data;
}
void	CharStrMap_free(CharStrMap_t *self, void (*fnFree)(void*))
{
	//return;
	if (!self)
		return;
	CharStrMap_iterate(self,CharStrMap_freeIter,fnFree);
	if (self->dict)
		KeyNode_free((KeyNode_t*)self->dict);
	if (self->nameBuffer) genfree(self->nameBuffer,self->maxNameLen);
	genfree(self,sizeof(CharStrMap_t));
}

void	CharStrMap_iterate(CharStrMap_t *self, void* (*fn)(void*,const char *,void*), void *data)
{
	if (self->dict)
		KeyNode_iterate((KeyNode_t*)self->dict,self->nameBuffer,self->nameBuffer,fn,data);
}

struct CharStrMapGet_s{
	void	*ret;
	int		cnt;
};

static void* CharStrMap_getNthCallback(void *fndata, const char *name, void *browsedata)
{
	struct CharStrMapGet_s*	search	= (struct CharStrMapGet_s*)fndata;

	if (search->cnt < 0) return fndata;
	if (search->cnt-- == 0) search->ret = browsedata;
	return fndata;
}
void*	CharStrMap_getNth(CharStrMap_t *self, uint n)
{
	struct CharStrMapGet_s	 search = {NULL,n};
	if (n>=self->entryCount) return NULL;
	CharStrMap_iterate(self,CharStrMap_getNthCallback,&search);
	return search.ret;
}

uint	CharStrMap_getCount(CharStrMapPTR	self){
	return self->entryCount;
	
}
