// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt

#include <luxcore/contvector.h>
#include <luxcore/memorygeneric.h>
#include <string.h>

void	ContVector_reserve(ContVector_t *cv,size_t n)
{
	const size_t 
		e = cv->elemsize,
		s = ContVector_size(cv),
		c = ContVector_capacity(cv);

	if (c >= n) return;
	if (c == 0)
	{
		if (cv->alignsize){
			cv->beg = (byte*)genzallocaligned(e * n, cv->alignsize);
		}
		else{
			cv->beg = (byte*)genzalloc(e * n);
		}
	}
	else
	{
		if (cv->alignsize){
			cv->beg = (byte*)genreallocaligned(cv->beg, n*e,  cv->alignsize, c*e);
		}
		else{
			cv->beg = (byte*)genrealloc(cv->beg, e * n, c*e);
		}
	}
	cv->end = cv->beg + s*e;
	cv->eos = cv->beg + n*e;
	LUX_ASSERT(ContVector_capacity(cv) >= n);
	LUX_ASSERT(ContVector_size(cv) == s);
}

void ContVector_prepGrowth(ContVector_t *cv, size_t delta)
{
	const size_t s = ContVector_size(cv);
	const size_t smallThreshold = CONTVECTOR_SMALL;
	if (s < smallThreshold)
	{
		ContVector_reserve(cv,MAX(smallThreshold, delta));
	}
	else
	{
		const size_t multiply = 3;
		const size_t divide = 2;
		const size_t suggestedSize = (s * multiply) / divide;
		ContVector_reserve(cv,MAX(s + delta, suggestedSize));
	}
}

ContVector_t ContVector_copy(ContVector_t* cv)
{
	ContVector_t	ncv = ContVector_new(cv->elemsize);
	ContVector_reserve(&ncv,ContVector_size(cv));
	memcpy(ncv.beg, cv->beg, ContVector_fullsize(cv));

	return ncv;
}

void	ContVector_pushBack(ContVector_t *cv, const void *data)
{
	if (cv->end == cv->eos) 
	{
		ContVector_prepGrowth(cv,1);
	}

	memcpy(cv->end,data,cv->elemsize);
}

void	ContVector_popBack(ContVector_t *cv)
{
	LUX_ASSERT(!ContVector_empty(cv));
	cv->end--;
}

void	ContVector_clear(ContVector_t *cv){
	if(!ContVector_empty(cv)){
		if (cv->alignsize){
			genfreealigned(cv->beg,ContVector_size(cv)*cv->elemsize);
		}
		else{
			genfree(cv->beg,ContVector_size(cv)*cv->elemsize);
		}

		cv->end = cv->beg;
		cv->eos = cv->beg;
	}
}

LUX_INLINE void	ContVector_insertMany(ContVector_t *cv, size_t idx, const void *data, size_t cnt)
{
	const size_t 
		e = cv->elemsize,
		s = ContVector_size(cv),
		c = ContVector_capacity(cv);

	const size_t
		bytes = e*cnt;

	if (s+cnt > c) 
	{
		ContVector_prepGrowth(cv,cnt);
	}

	{
		byte*	idxpos = (byte*)ContVector_at(cv,idx);

		if (idx < s){
			memmove(idxpos+bytes,idxpos,(s-idx)*e);
		}
		memcpy(idxpos,data,bytes);
		cv->end += (e*cnt);
	}
}

void	ContVector_insert(ContVector_t *cv, size_t idx, const void *data)
{
	ContVector_insertMany(cv,idx,data,1);
}

void	ContVector_insertRepeat(ContVector_t *cv, size_t idx, void *data, size_t cnt)
{
	const size_t 
		e = cv->elemsize,
		s = ContVector_size(cv),
		c = ContVector_capacity(cv);

	const size_t
		bytes = e*cnt;

	if (s+cnt > c) 
	{
		ContVector_prepGrowth(cv,cnt);
	}

	{
		byte*	idxpos = (byte*)ContVector_at(cv,idx);
		byte*	endpos = idxpos+bytes;
		
		if (idx < s){
			memmove(idxpos+bytes,idxpos,(s-idx)*e);
		}
		for (; idxpos < endpos; idxpos += e){
			memcpy(idxpos,data,e);
		}
		
		cv->end = idxpos;
	}
}

LUX_INLINE void	ContVector_removeMany(ContVector_t *cv, size_t idx, size_t cnt)
{
	const size_t 
		e = cv->elemsize,
		s = ContVector_size(cv),
		c = ContVector_capacity(cv);

	const size_t
		ncnt = MIN(cnt,(s-idx));

	const size_t
		bytes = e*ncnt;

	byte*	idxpos = (byte*)ContVector_at(cv,idx);
	byte*	frompos = idxpos+bytes;

	if (frompos < cv->end){
		memmove(idxpos,frompos,bytes);
	}

	cv->end -= bytes;
}

void	ContVector_remove(ContVector_t *cv, size_t idx)
{
	ContVector_removeMany(cv,idx,1);
}



void	ContVector_shrink(ContVector_t *cv)
{
	const size_t 
		e = cv->elemsize,
		s = ContVector_size(cv),
		c = ContVector_capacity(cv);

	if (s == 0) return;

	if (cv->alignsize){
		cv->beg = (byte*)genreallocaligned(cv->beg, s*e, cv->alignsize, c*e);
	}
	else{
		cv->beg = (byte*)genrealloc(cv->beg,s * e, c*e);
	}

	cv->end = cv->beg + s*e;
	cv->eos = cv->end;
	LUX_ASSERT(ContVector_capacity(cv) >= s);
	LUX_ASSERT(ContVector_size(cv) == s);
}

void	ContVector_resize(ContVector_t *cv,size_t cnt, const void *fill)
{
	const size_t 
		e = cv->elemsize,
		s = ContVector_size(cv),
		c = ContVector_capacity(cv);

	if (c == cnt) return;

	ContVector_reserve(cv,cnt);
	if (cnt > s){
		byte* curend = cv->end;
		byte* end = cv->eos;
		const byte* LUX_RESTRICT from = (const byte*)fill;

		for (; curend < end; curend += e){
			memcpy(curend,from,e);
		}
		cv->end = end;
	}
}
