// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt

#include <luxcore/memorypool.h>
#include <luxcore/debug.h>

#include <string.h>

//////////////////////////////////////////////////////////////////////////
// MemoryPool
typedef struct MemoryPool_s{
	const char		*name;
	char			*dataorig;	// the allocated memory
	char			*data;		// the aligned start
	size_t			inuse;		// what is used in bytes
	size_t			total;		// the total in bytes
	int				resized;
}MemoryPool_t;

MemoryPool_t *MemoryPool_new(const char* name, size_t totalbytes)
{
	MemoryPool_t *mem;

	//if (l_poolMutex == -1) l_poolMutex = critical_create();

	mem = (MemoryPool_t*)MEMORY_GLOBAL_MALLOC(sizeof(MemoryPool_t)+strlen(name)+1);
	memset(mem,0,sizeof(MemoryPool_t));

	mem->name = (const char*)(mem+1);
	strcpy((char*)mem->name,name);

	MemoryPool_initMin(mem,totalbytes);

	return mem;
}
void* MemoryPool_zalloc(MemoryPool_t *mem, size_t size)
{
	char *pMem;
	// always pad size to 4-byte alignment
#ifdef LUX_MEMORY_POOL_PAD4
	size_t padsize = size%4;
	size += padsize ? 4-padsize : 0;
#endif

	if (size+mem->inuse > mem->total){
		LUX_PRINTF("ERROR: MemPoolMalloc failed - %s - size: %d inuse: %d total: %d\n",mem->name,size,mem->inuse,mem->total);
		LUX_ASSERT(0);
		return NULL;
	}
	else{
		pMem = &mem->data[mem->inuse];
		mem->inuse += size;
		memset(pMem,0,size);
		return pMem;
	}
}
void* MemoryPool_zallocAligned(MemoryPool_t *mem, size_t size, size_t align)
{
	char *pMem;

	size_t padsize;
	size_t offset;

	// shift mem->inuse
	offset = ((size_t)(&mem->data[mem->inuse]))%align;
	offset = offset ? align-offset : 0;
	mem->inuse += offset;

	// add pad to 4
	padsize = size%4;
	size += padsize ? 4-padsize : 0;

	if (size+mem->inuse > mem->total){
		LUX_PRINTF("ERROR: MemPoolMalloc failed - %s - size: %d inuse: %d total: %d\n",mem->name,size,mem->inuse,mem->total);
		LUX_ASSERT(0);
		return NULL;
	}
	else{
		pMem = &mem->data[mem->inuse];
		mem->inuse += size;
		memset(pMem,0,size);
		return pMem;
	}
}
void* MemoryPool_malloc(MemoryPool_t *mem, size_t size)
{
	char *pMem;

	if (size+mem->inuse > mem->total){
		LUX_PRINTF("ERROR: MemPoolMalloc failed - %s - size: %d inuse: %d total: %d\n",mem->name,size,mem->inuse,mem->total);
		LUX_ASSERT(0);
		return NULL;
	}
	else{
		pMem = &mem->data[mem->inuse];
		mem->inuse += size;
		return pMem;
	}
}
void* MemoryPool_current(MemoryPool_t *mem)
{
	void *ptr;

	ptr = &mem->data[mem->inuse];

	return ptr;
}
// returns true if ptr is inside pool
booln	 MemoryPool_checkPtr(MemoryPool_t *mem, void *ptr)
{
	if ((char*)ptr >= mem->data && (char*)ptr < mem->data+mem->total)
		return TRUE;
	LUX_PRINTF("ERROR: MemPoolPtr failed - %s - ptr: %p inuse: %d total: %d\n",mem->name,ptr,mem->inuse,mem->total);
	LUX_ASSERT(0);
	return FALSE;
}

// from this point on new allocs will be done
// if outside pool false is returned
void MemoryPool_setBegin(MemoryPool_t *mem, void *ptr){

	if (!MemoryPool_checkPtr(mem,ptr)){
		LUX_PRINTF("ERROR: MemPoolSetBegin failed - %s - ptr: %d inuse: %d total: %d\n",mem->name,ptr,mem->inuse,mem->total);
		LUX_ASSERT(0);
		return;
	}

	mem->inuse = (int)((char*)ptr - mem->data);
}
int MemoryPool_initMin(MemoryPool_t *mem, size_t totalbytes)
{
	char *newdata = mem->dataorig;

	if (mem->dataorig){
		if (totalbytes > mem->total){
			mem->dataorig = newdata = (char*)MEMORY_GLOBAL_REALLOC(mem->data, totalbytes + sizeof(uint32));
			mem->data = (char*)pointerAlign(mem->dataorig,sizeof(uint32));
			mem->total = totalbytes;
		}
	}
	else{
		mem->dataorig = (char*)MEMORY_GLOBAL_MALLOC(totalbytes + sizeof(uint32));
		mem->data = (char*)pointerAlign(mem->dataorig,sizeof(uint32));
		mem->total = totalbytes;
	}

	mem->inuse = 0;
	mem->resized = (newdata != mem->dataorig);

	return mem->resized;
}

int	 MemoryPool_popResized(MemoryPool_t *mem)
{
	int resized = mem->resized;
	mem->resized = FALSE;

	return resized;
}

void MemoryPool_clear(MemoryPool_t *mem)
{
	mem->inuse = 0;
}

void MemoryPool_free(MemoryPool_t *mem)
{
	if (mem->dataorig != NULL)
		MEMORY_GLOBAL_FREE(mem->dataorig);

	mem->data = NULL;
	mem->inuse = 0;
	mem->total = 0;
}

size_t	MemoryPool_bytesLeft(MemoryPool_t *mem)
{
	return (mem->total-mem->inuse);
}
size_t MemoryPool_bytesUsed(MemoryPool_t *mem)
{
	return (mem->inuse);
}
size_t MemoryPool_bytesTotal(MemoryPool_t *mem)
{
	return (mem->total);
}
