// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt

#include <luxcore/refsys.h>
#include <luxcore/contvector.h>
#include <luxcore/memorygeneric.h>

//////////////////////////////////////////////////////////////////////////
#define OBJREFSYS_ALLOC_PERPAGE	((512-(sizeof(ObjRef_t)))/(sizeof(ObjRef_t)))

typedef struct ObjRefPage_s{
	ObjRef_t				next;	// sacrifice one for pagelist and to keep alignment well
	ObjRef_t				block[OBJREFSYS_ALLOC_PERPAGE];
}ObjRefPage_t;

//////////////////////////////////////////////////////////////////////////

ObjRefSysPTR	ObjRefSys_new()
{
	ObjRefSysPTR refsys = (ObjRefSysPTR)genmalloc(sizeof(ObjRefSys_t));
	ObjRefSys_init(refsys);
	return refsys;
}
void			ObjRefSys_delete(ObjRefSysPTR sys)
{
	ObjRefSys_deinit(sys,NULL);
	genfree(sys,sizeof(ObjRefSys_t));
}

void ObjRefSys_popNoDelete(ObjRefSysPTR sys)
{
	if (--sys->nodelete == 0){
		while (ContVector_size(&sys->nodelvec)){
			ObjRef_t* cref = (ObjRef_t*)ContVector_back(&sys->nodelvec);

			// in the ugly case someone has created a new reference
			// within the NoDelete section
			if (cref->usecounter == 0){
				ObjRefSys_deleteRef(sys,cref);
			}
			

			ContVector_popBack(&sys->nodelvec);
		}
	}
	LUX_ASSERT(sys->nodelete >= 0);
}

void ObjRefSys_pushNoDelete(ObjRefSysPTR sys)
{
	LUX_ASSERT(sys->nodelete >= 0);

	sys->nodelete++;
}


void ObjRefSys_init(ObjRefSysPTR sys)
{
	memset(sys,0,sizeof(ObjRefSys_t));
	sys->allocator.lastinpage = OBJREFSYS_ALLOC_PERPAGE;
	ContVector_init(&sys->typeinfovec,sizeof(ObjTypeInfo_t));
	ContVector_init(&sys->nodelvec,sizeof(ObjRefPTR));
}

void ObjRefSys_deinit(ObjRefSysPTR sys, ObjRefCheckDeleteFn *func)
{
	// cleanup all refs
	ObjRefPage_t*	pagelist = sys->allocator.pagelist;

	while (sys->nodelete > 0){
		ObjRefSys_popNoDelete(sys);
	}

	while (pagelist){
		ObjRefPage_t*	page = pagelist;
		uint32		i;

		pagelist = (ObjRefPage_t*)page->next.id.ptr;

		for (i = 0; i < OBJREFSYS_ALLOC_PERPAGE; i++)
		{
			ObjRef_t	*cref = &page->block[i];
			if (cref->id.type >= OBJREF_TYPE_USERSTART &&
				(!func || func(cref)) &&
				sys->typeinfos[cref->id.type].fnDelete)
			{
				sys->typeinfos[cref->id.type].fnDelete(cref);
			}
		}

		genfree(page,512);
	}

	ContVector_clear(&sys->typeinfovec);
	ContVector_clear(&sys->nodelvec);
	memset(sys,0,sizeof(ObjRefSys_t));
}

void ObjRefSys_register(ObjRefSys_t *sys, ObjRefType_t type,ObjTypeInfo_t infocpy)
{
	static const ObjTypeInfo_t nullinfo = {NULL};
	size_t sz = ContVector_size(&sys->typeinfovec);

	LUX_ASSERT(type != OBJREF_TYPE_FREEALLOC);

	if ((size_t)type+1 > sz){
		ContVector_resize(&sys->typeinfovec,(type+1),&nullinfo);
		sys->typeinfos = (ObjTypeInfo_t*) ContVector_front(&sys->typeinfovec);
	}
	sys->typeinfos[type] = infocpy;
}

void ObjRefSys_deleteAlloc(ObjRefSys_t *sys,ObjRef_t *data)
{
	ObjRefAllocSys_t*	allocator = &sys->allocator;

	if (data){
		data->id.ptr = allocator->freerefs;
		data->id.type = OBJREF_TYPE_FREEALLOC;
		allocator->freerefs = data;
		LUX_PROFILING_OP(sys->count--);
	}
}

static ObjRef_t* ObjRefSys_newAlloc(ObjRefSys_t *sys){
	ObjRefAllocSys_t*	allocator = &sys->allocator;
	ObjRef_t* block;

	LUX_PROFILING_OP(sys->count++);
	if ( block = allocator->freerefs){
		allocator->freerefs = (ObjRef_t*)block->id.ptr;
		return block;
	}
	else{
		if (allocator->lastinpage == OBJREFSYS_ALLOC_PERPAGE){
			ObjRefPage_t *page = (ObjRefPage_t*)genzalloc(512);
			LUX_ASSERT(sizeof(ObjRefPage_t) <= 512);
			page->next.id.ptr = allocator->pagelist;			
			allocator->pagelist = page;
			allocator->lastinpage = 0;
		}
		return &allocator->pagelist->block[allocator->lastinpage++];
	}
}

//////////////////////////////////////////////////////////////////////////


ObjRefPTR	ObjRefSys_newRef(ObjRefSys_t *sys, ObjRefType_t type, void *ptr)
{
	ObjRefPTR	cref = (ObjRefPTR) ObjRefSys_newAlloc(sys);

	LUX_ASSERT(type >= OBJREF_TYPE_USERSTART);

	cref->id.ptr = ptr;
	cref->id.type = type;
	cref->usecounter = 1;
	cref->weakcounter = 1;

	return cref;
}


void  ObjRefSys_deleteRef(ObjRefSys_t *sys, ObjRefPTR cref)
{
	if (!sys->nodelete){
		if (ObjRef_getId(cref)){
			ObjRefDeleteFn*	fndel = sys->typeinfos[cref->id.type].fnDelete;

			if (fndel){
				fndel(cref);
			}
		}
		cref->id.type = OBJREF_TYPE_DELETED;
		cref->id.ptr = NULL;
		ObjRefSys_releaseWeakRef(sys,cref);
	}
	else{
		ContVector_pushBack(&sys->nodelvec,cref);
	}
}

booln ObjRefSys_invalidateRef(ObjRefSysPTR sys,ObjRefPTR cref, booln calldestructor)
{

	if (ObjRef_getId(cref) && calldestructor){
		ObjRefDeleteFn*	fndel = sys->typeinfos[cref->id.type].fnDelete;

		if (fndel){
			fndel(cref);
		}
	}
	cref->id.type = OBJREF_TYPE_DELETED;
	cref->id.ptr = NULL;
	return (cref->usecounter > 0);
}
