// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt
//
// <Author: Christoph Kubisch>
// Based on Nvidia SDK


#ifndef __LUXMATH_FASTMATH_H__
#define __LUXMATH_FASTMATH_H__

#include <luxmath/basetypes.h>

#ifdef __cplusplus
extern "C"{
#endif

#define MATH_MAX_CIRCLE_ANGLE	8192

typedef struct FastMathCache_s{
	unsigned int	fast_sqrt_table[0x10000];
	float			mathcossintable[MATH_MAX_CIRCLE_ANGLE];
	float			float16conv[0x10000];
}FastMathCache_t;


// initializes lookup tables
void fastmath_init(FastMathCache_t *cache);
// assumes lookup tables already inited
void fastmath_initset(FastMathCache_t *cache);

void float2int(int *int_pointer,const float f);
float fastposmod(float in, float mod);

// using lookup tables
	// tablesize (256 KB)
float fastsqrt(float n);
	// tablesize for both (32 KB)
float fastcos(const float n);
float fastsin(const float n);
	// tablesize (256 KB)
float fastfloat16to32(const float16 n);

	// less accurate
float fastsqrtinv(float n);

int   fastiround(float f);
float fastfrac(float f);
int   fastifloor(float f);

// macros
//#define FP_INV(r,p)
//#define FP_EXP(e,p)
//#define FP_NORM_TO_BYTE(i,p)

//#define FP_INTCAST(f)
//#define FP_UINTCAST(f)

//#define FP_GREATER_ZERO(f)
//#define FP_LESS_ZERO(f)
//#define FP_GEQUAL_ZERO(f)
//#define FP_LEQUAL_ZERO(f)

//#define FP_LESS_ONE(f)
//#define FP_GREATER_ONE(f)
//#define FP_LEQUAL_ONE(f)
//#define FP_GEQUAL_ONE(f)

//////////////////////////////////////////////////////////////////////////

#define FP_INTCAST(f)	(*((int *)(&f)))
#define FP_UINTCAST(f)	(*((unsigned int *)(&f)))

#define FP_GEQUAL_ZERO(f)	((FP_UINTCAST(f) <= 0x80000000U))
#define FP_LESS_ZERO(f)		((FP_UINTCAST(f) > 0x80000000U))
#define FP_GREATER_ZERO(f)	((FP_INTCAST(f) > 0))
#define FP_LEQUAL_ZERO(f)	((FP_INTCAST(f) <= 0))


#define FP_LESS_ONE(f)		((FP_INTCAST(f) < 0x3F800000))
#define FP_GREATER_ONE(f)	((FP_INTCAST(f) > 0x3F800000))
#define FP_LEQUAL_ONE(f)	((FP_INTCAST(f) <= 0x3F800000))
#define FP_GEQUAL_ONE(f)	((FP_INTCAST(f) >= 0x3F800000))

#define FP_BITS(fp) (*(uint32 *)&(fp))
#define FP_ABS_BITS(fp) (FP_BITS(fp)&0x7FFFFFFF)
#define FP_SIGN_BIT(fp) (FP_BITS(fp)&0x80000000)
#define FP_ONE_BITS 0x3F800000

#if defined(_DEBUG)
LUX_INLINE void float2int(int *int_pointer,const float flt){
	*(int_pointer)=(int)((flt) + 0.5f);
}
#else
#if defined(LUX_COMPILER_MSC) && defined(LUX_ARCH_X86)
LUX_INLINE void float2int(int *int_pointer,const float f)
{
	__asm  fld  f
	__asm  mov  edx,int_pointer
	__asm  FRNDINT
	__asm  fistp dword ptr [edx];

}
#else
LUX_INLINE void float2int(int *int_pointer,const float flt)
{
	*(int_pointer)=(int)((flt) + 0.5f);
}
#endif
#endif

LUX_INLINE float  fastposmod(float in, float mod){
	return ((FP_LESS_ZERO(in)) ? ((mod) - (float)fmod(-(in),(mod)))  : (float)fmod((in),(mod)));
}

// 1/sqrt
LUX_INLINE float fastsqrtinv (float x){
	float xhalf = 0.5f*x;
	int i = *(int*)&x;
	i = 0x5f3759df - (i>>1);
	x = *(float*)&i;
	x = x*(1.5f - xhalf*x*x);
	return x;
}

//////////////////////////////////////////////////////////////////////////
// misc

#define FP_INV(r,p)                                                          \
{                                                                            \
	int _i = 2 * FP_ONE_BITS - *(int *)&(p);                                 \
	r = *(float *)&_i;                                                       \
	r = r * (2.0f - (p) * r);                                                \
}


#define FP_EXP(e,p)                                                          \
{                                                                            \
	int _i;                                                                  \
	e = -1.44269504f * (float)0x00800000 * (p);                              \
	_i = (int)e + 0x3F800000;                                                \
	e = *(float *)&_i;                                                       \
}


#if defined(LUX_COMPILER_MSC)
#define FP_NORM_TO_BYTE(i,p)						\
{													\
	float _temp = (p) + 1.0f;						\
	(i) = (*(int *)&_temp >= 0x40000000)?0xFF:		\
	(*(int *)&_temp <=0x3F800000)?0:				\
	((*(int *)&_temp) >> 15)&0xFF;					\
}
#else // if buggy compiler (like gcc...)
#define FP_NORM_TO_BYTE(i,p)										\
{																	\
	(i) = ((p)>=1.0f)?0xFF:((p)<=0)?0:(unsigned char)((p)*255.0f);	\
}
#endif

//////////////////////////////////////////////////////////////////////////
// rounding and fractional

#if !defined(_DEBUG) && defined(LUX_COMPILER_GCC) && defined(LUX_ARCH_X86)
LUX_INLINE int fastiround(float f)
{
	int r;
	__asm__ ("fistpl %0" : "=m" (r) : "t" (f) : "st");
	return r;
}
#elif !defined(_DEBUG) && defined(LUX_COMPILER_MSC) && defined(LUX_ARCH_X86)
LUX_INLINE int fastiround(float f)
{
	int r;
	_asm {
		fld f
			fistp r
	}
	return r;
}
#else
LUX_INLINE int fastiround(float f)
{
	return ((int) (((f) >= 0.0F) ? ((f) + 0.5F) : ((f) - 0.5F)));
}
#endif

LUX_INLINE int fastifloor(float f)
{
	int i = fastiround(f);
	return (i > f) ? i - 1 : i;
}
LUX_INLINE float fastfrac(float f)
{
	return ((f) - fastifloor(f));
}

//////////////////////////////////////////////////////////////////////////
extern FastMathCache_t*	g_fmcache;

LUX_INLINE float fastsqrt(float n)
{

	if (FP_BITS(n) == 0)
		return 0.0;                 // check for square root of 0

	FP_BITS(n) = g_fmcache->fast_sqrt_table[(FP_BITS(n) >> 8) & 0xFFFF] | ((((FP_BITS(n) - 0x3F800000) >> 1) + 0x3F800000) & 0x7F800000);

	return n;
}

//////////////////////////////////////////////////////////////////////////
// sin cos
#define MATH_HALF_MAX_CIRCLE_ANGLE		(MATH_MAX_CIRCLE_ANGLE/2)
#define MATH_QUARTER_MAX_CIRCLE_ANGLE	(MATH_MAX_CIRCLE_ANGLE/4)
#define MATH_MASK_MAX_CIRCLE_ANGLE		(MATH_MAX_CIRCLE_ANGLE - 1)


// from flipcode
LUX_INLINE float fastcos(const float n)
{
	float f;
	int i;

	f = n * MATH_HALF_MAX_CIRCLE_ANGLE * DIV_PI;
	float2int(&i, f);
	if (i < 0)
	{
		return g_fmcache->mathcossintable[((-i) + MATH_QUARTER_MAX_CIRCLE_ANGLE)&MATH_MASK_MAX_CIRCLE_ANGLE];
	}
	else
	{
		return g_fmcache->mathcossintable[(i + MATH_QUARTER_MAX_CIRCLE_ANGLE)&MATH_MASK_MAX_CIRCLE_ANGLE];
	}
}

LUX_INLINE float fastsin(const float n)
{
	float f;
	int i;

	f = n * MATH_HALF_MAX_CIRCLE_ANGLE * DIV_PI;
	float2int(&i, f);
	if (i < 0)
	{
		return g_fmcache->mathcossintable[(-((-i)&MATH_MASK_MAX_CIRCLE_ANGLE)) + MATH_MAX_CIRCLE_ANGLE];
	}
	else
	{
		return g_fmcache->mathcossintable[i&MATH_MASK_MAX_CIRCLE_ANGLE];
	}
}

LUX_INLINE float fastfloat16to32(const float16 n)
{
	return g_fmcache->float16conv[(ushort)n];
}


#ifdef __cplusplus
}
#endif

#endif
