// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt
//
// <Author: Christoph Kubisch>


#ifndef __LUXMATH_FRUSTUM_H__
#define __LUXMATH_FRUSTUM_H__

#include <luxmath/basetypes.h>

#ifdef __cplusplus
extern "C"{
#endif

//////////////////////////////////////////////////////////////////////////
// Frustum

void Frustum_update(Frustum_t *frustum, const Matrix44 viewproj);

// returns FALSE if inside else TRUE
// the plane it failed on will be saved to it as well
booln  Frustum_checkPointCoherent(const Frustum_t *frustum, const Vector3 vec, int *clipplane);

// returns FALSE if inside/intersect
booln  Frustum_checkSphereCoherent(const Frustum_t *frustum, const Vector3 center, const float radius,int *clipplane);

// returns FALSE if sphere is fully inside
booln  Frustum_checkSphereFull(const Frustum_t *frustum, const Vector3 center, const float radius);

// returns FALSE if inside/intersect
booln  Frustum_checkAABBvCoherent(const Frustum_t *frustum, const float minmaxs[6], int *clipplane);

// masking & temporal coherency
// cullF can report outside even if extends of box are outside all planes
CullType_t  Frustum_cullAABBvMaskedCoherent(const Frustum_t *pFrustum, const float minmax[6], int in_mask, int *out_mask, int *inoutstart_id);

CullType_t	Frustum_cullPoints(const Frustum_t *frustum, const Vector4 *vecarray, const int numVec);
CullType_t  Frustum_cullAABBv(const Frustum_t *frustum, const float minmaxs[6]);

// generates corner points for a frustum
void Frustum_getCorners(const Frustum_t *frustum, Vector3 box[FRUSTUM_CORNERS]);
void Frustum_fromCorners(Frustum_t *frustum, const Vector3 box[FRUSTUM_CORNERS]);

void Frustum_updateSigns(Frustum_t *frustum);

//////////////////////////////////////////////////////////////////////////
//

LUX_INLINE booln Frustum_checkPointCoherent(const Frustum_t *pFrustum,const Vector3 vec, int *clipplane)
{
	int i,skip;

	skip = (*clipplane);
	if(pFrustum->fplanes[skip].pvec[0] * vec[0] + pFrustum->fplanes[skip].pvec[1] * vec[1] + pFrustum->fplanes[skip].pvec[2] * vec[2] + pFrustum->fplanes[skip].pvec[3] <= 0)
	{
		return TRUE;
	}
	for(i = 0; i < FRUSTUM_PLANES; i++ )
	{
		if(i != skip && pFrustum->fplanes[i].pvec[0] * vec[0] + pFrustum->fplanes[i].pvec[1] * vec[1] + pFrustum->fplanes[i].pvec[2] * vec[2] + pFrustum->fplanes[i].pvec[3] <= 0)
		{
			*clipplane = i;
			return TRUE;
		}
	}

	return FALSE;
}

LUX_INLINE booln Frustum_checkSphereCoherent(const Frustum_t *pFrustum,const Vector3 vec,const float radius, int *clipplane)
{
	int i,skip;

	skip = *clipplane;
	if(pFrustum->fplanes[skip].pvec[0] * vec[0] + pFrustum->fplanes[skip].pvec[1] * vec[1] + pFrustum->fplanes[skip].pvec[2] * vec[2] + pFrustum->fplanes[skip].pvec[3] <= -radius)
	{
		return TRUE;
	}

	for( i = 0; i < FRUSTUM_PLANES; i++ )
	{
		if(i != skip &&  pFrustum->fplanes[i].pvec[0] * vec[0] + pFrustum->fplanes[i].pvec[1] * vec[1] + pFrustum->fplanes[i].pvec[2] * vec[2] + pFrustum->fplanes[i].pvec[3] <= -radius )
		{
			*clipplane = i;
			return TRUE;
		}
	}
	return FALSE;
}


LUX_INLINE booln 	Frustum_checkAABBvCoherent(const Frustum_t *pFrustum,const float minmax[6], int *clipplane)
{
	int i,skip;
	skip = *clipplane;
	{
		const FrustumPlane_t *sp = &pFrustum->fplanes[skip];
		if ((sp->pvec[0] * minmax[sp->px]) + 
			(sp->pvec[1] * minmax[sp->py]) + 
			(sp->pvec[2] * minmax[sp->pz]) + sp->pvec[3] < 0)
		{
			return TRUE;
		}
	}


	for (i = 0; i < FRUSTUM_PLANES; i++){
		if (i != skip){

			const FrustumPlane_t *sp = &pFrustum->fplanes[i];
			if ((sp->pvec[0] * minmax[sp->px]) + 
				(sp->pvec[1] * minmax[sp->py]) + 
				(sp->pvec[2] * minmax[sp->pz]) + sp->pvec[3] < 0)
			{
				*clipplane = i;
				return TRUE;
			}
		}
	}

	return FALSE;
}

#ifdef __cplusplus
}
#endif

#endif

