// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt
//
// <Author: Christoph Kubisch>


#ifndef __LUXMATH_MATRIX44_H__
#define __LUXMATH_MATRIX44_H__

#include <luxmath/basetypes.h>
#include <luxmath/fastmath.h>
#include <luxmath/vector3.h>

#ifdef __cplusplus
extern "C"{
#endif

const float *Matrix44GetIdentity();
void Matrix44Identity(Matrix44 m_mat);
void Matrix44Copy(Matrix44 dst, const Matrix44 src);
void Matrix44CopyRot(Matrix44 dst, const Matrix44 src);
void Matrix44CopyRotTransposed(Matrix44 dst, const Matrix44 src);

void Matrix44SetTranslation( Matrix44 m_mat, const Vector3 translation );
void Matrix44SetInvTranslation( Matrix44 m_mat, const Vector3 translation );
void Matrix44SetScale( Matrix44 mat, const Vector3 scale);
void Matrix44SetRotRows(Matrix44 dst, 
						float a, float b, float c, 
						float d, float e, float f, 
						float g, float h, float i);

float*	Matrix44GetTranslation(const Matrix44 mat, Vector3 vec);

void Matrix44Clear(Matrix44 b);
void Matrix44ClearRot(Matrix44 b);
void Matrix44ClearTranslation(Matrix44 matrix);

void Matrix44VectorTransform( const Matrix44 mat, Vector3 v1 );
void Matrix44VectorInvTransform( const Matrix44 m_mat, Vector3 pVect );
void Matrix44VectorRotate( const Matrix44 mat, Vector3 v1 );
void Matrix44VectorInvRotate( const Matrix44 m_mat, Vector3 pVect );
void Matrix44VectorTranslate( const Matrix44 m_mat, Vector3 pVect );
void Matrix44VectorInvTranslate( const Matrix44 m_mat, Vector3 pVect );

void Matrix44Multiply( Matrix44 newmat, const Matrix44 m_mat,  const Matrix44 mat );
void Matrix44Multiply1( Matrix44 m_mat,  const Matrix44 mat );
void Matrix44Multiply2( const Matrix44 m_mat,  Matrix44 mat );
void Matrix44MultiplyFull( Matrix44 clip, const Matrix44 proj , const Matrix44 modl);

void Matrix44MultiplyRot(Matrix44 newmat, const Matrix44 m_mat, const Matrix44 mat );
void Matrix44MultiplyRot1( Matrix44 m_mat, const Matrix44 mat );
void Matrix44MultiplyRot2( const Matrix44 m_mat,  Matrix44 mat );

void Matrix44Transpose(Matrix44 mat, const Matrix44 oldmat);
void Matrix44Transpose1(Matrix44 mat);
void Matrix44TransposeRot(Matrix44 mat, const Matrix44 oldmat);
void Matrix44TransposeRot1(Matrix44 mat);
void Matrix44TransposeRotIdentity(Matrix44 mat, const Matrix44 oldmat);

void Matrix44Invert(Matrix44 inverse, const Matrix44 src);
void Matrix44Invert1(Matrix44 m);
void Matrix44AffineInvert(Matrix44 outmat, const Matrix44 mat);

// rotation
void Matrix44Aim(Matrix44 mat, Vector3 forward,Vector3 up, int axis);
void Matrix44LookAt(Matrix44 mat, Vector3 from, Vector3 to, Vector3 upn);

void Matrix44FromEulerZYX( Matrix44 m_mat, const Vector3 angles );
void Matrix44FromEulerZYXdeg( Matrix44 m_mat, const Vector3 angles );
void Matrix44FromEulerXYZ(Matrix44 m_mat, Vector3 angles);
void Matrix44FromAngleAxisFast(Matrix44 mat, const Vector3 axis, 
							   float cos, float sin, float oneminuscos);
void Matrix44FromAngleAxis(Matrix44 matrix, float anglerad, const Vector3 axis);
void Matrix44RotateAngle(Matrix44 mat, Vector3 from, Vector3 to);
void Matrix44RotateAroundVector(Matrix44 mat, Vector3 axis, float angleRad);
void Matrix44RotateAroundPoint( Matrix44 transforms,  const Vector3 center, const Vector3 angles);

void Matrix44ToEulerXYZ( const Matrix44 a_mat, Vector3 angles);
void Matrix44ToEulerZYX( const Matrix44 a_mat, Vector3 angles);

// projection
void Matrix44Perspective(Matrix44 mat, const float fov, 
					const float front, const float back, const float aspect);
void Matrix44PerspectiveInf(Matrix44 mat, const float fov, const float front, const float aspect);
void Matrix44Ortho(Matrix44 mat, const float height, 
					const float front, const float back, const float aspect);
void Matrix44OrthoDirect(Matrix44 mat, const float left, const float right, 
					const float	bottom, const float top, 
					const float front, const float back);
void Matrix44PlaneProjection(Matrix44 matrix, const Vector3 planenormal);
void Matrix44ModifyProjectionClipplane(Matrix44 projmatrix,const Matrix44 mview, 
					const Matrix44 mviewinv, const Vector4 clipPlane);
void Matrix44Reflection(Matrix44 matrix,Vector4 plane);

// misc
float Matrix44CompareRot(Matrix44 a, Matrix44 b);
void  Matrix44Swizzle1(Matrix44 mat, uint axis[3], Vector3 dirs);



//////////////////////////////////////////////////////////////////////////

#ifdef LUX_COMPILER_MSC
#pragma warning( push )
#pragma warning( disable : 4028 )
#endif

static const LUX_ALIGNSIMD_V(float _matrix44_ident[16]) =
{
	1.0f, 0.0f, 0.0f, 0.0f,
	0.0f, 1.0f, 0.0f, 0.0f,
	0.0f, 0.0f, 1.0f, 0.0f,
	0.0f, 0.0f, 0.0f, 1.0f,
};

LUX_INLINE void Matrix44Copy(Matrix44 dst, const float* LUX_RESTRICT src)
{
	(dst)[0]  = (src)[0];  (dst)[1]  = (src)[1];  
	(dst)[2]  = (src)[2];  (dst)[3]  = (src)[3];  
	(dst)[4]  = (src)[4];  (dst)[5]  = (src)[5];  
	(dst)[6]  = (src)[6];  (dst)[7]  = (src)[7];  
	(dst)[8]  = (src)[8];  (dst)[9]  = (src)[9];  
	(dst)[10] = (src)[10]; (dst)[11] = (src)[11]; 
	(dst)[12] = (src)[12]; (dst)[13] = (src)[13]; 
	(dst)[14] = (src)[14]; (dst)[15] = (src)[15];
}
LUX_INLINE void Matrix44CopyRot(Matrix44 dst, const float* LUX_RESTRICT src) 
{
	(dst)[0]  = (src)[0];  (dst)[1]  = (src)[1];  
	(dst)[2]  = (src)[2];   
	(dst)[4]  = (src)[4];  (dst)[5]  = (src)[5];  
	(dst)[6]  = (src)[6];    
	(dst)[8]  = (src)[8];  (dst)[9]  = (src)[9];  
	(dst)[10] = (src)[10];
}

LUX_INLINE void Matrix44CopyRotTransposed(Matrix44 dst, const float* LUX_RESTRICT src) 
{	
	(dst)[0]  = (src)[0];  (dst)[4]  = (src)[1];  
	(dst)[8]  = (src)[2];   
	(dst)[1]  = (src)[4];  (dst)[5]  = (src)[5];  
	(dst)[9]  = (src)[6];    
	(dst)[2]  = (src)[8];  (dst)[6]  = (src)[9];  
	(dst)[10] = (src)[10];
}

LUX_INLINE void Matrix44SetRotRows(Matrix44 dst, 
									float a, float b, float c, 
									float d, float e, float f, 
									float g, float h, float i)
{
	(dst)[0] =  (a);(dst)[1]  = (d); (dst)[2]  = (g);
	(dst)[4]  = (b);(dst)[5]  = (e); (dst)[6]  = (h);
	(dst)[8]  = (c);(dst)[9]  = (f); (dst)[10] = (i);
}

LUX_INLINE void Matrix44Clear(Matrix44 b)	
{
	(b)[0]  = (b)[1]  = (b)[2]  = (b)[3]  = 
	(b)[4]  = (b)[5]  = (b)[6]  = (b)[7]  = 
	(b)[8]  = (b)[9]  = (b)[10] = (b)[11] = 
	(b)[12] = (b)[13] = (b)[14] = (b)[15] = 0.0f;
}

LUX_INLINE void Matrix44ClearRot(Matrix44 b)	
{
	(b)[1] = (b)[2]  =   
	(b)[4] = (b)[6]  =   
	(b)[8] = (b)[9]  = 0.0f; 
	(b)[0] = (b)[5] = (b)[10] = 1.0f;
}

LUX_INLINE void Matrix44ClearTranslation(Matrix44 matrix) 
{
	matrix[12] = matrix[13] = matrix[14] = 0.0f;
}

LUX_INLINE void Matrix44FromAngleAxisFast(Matrix44 mat, const Vector3 axis, float c, float s, float oneminusc)
{
	float tmp1,tmp2;

	(mat)[0] = c + (axis)[0]*(axis)[0]*oneminusc;
	(mat)[5] = c + (axis)[1]*(axis)[1]*oneminusc;
	(mat)[10] = c + (axis)[2]*(axis)[2]*oneminusc;
	
	tmp1 = (axis)[0]*(axis)[1]*oneminusc;
	tmp2 = (axis)[2]*s;
	(mat)[1] = tmp1 + tmp2;
	(mat)[4] = tmp1 - tmp2;
	
	tmp1 = (axis)[0]*(axis)[2]*oneminusc;
	tmp2 = (axis)[1]*s;
	(mat)[2] = tmp1 - tmp2;
	(mat)[8] = tmp1 + tmp2;
	
	tmp1 = (axis)[1]*(axis)[2]*oneminusc;
	tmp2 = (axis)[0]*s;
	(mat)[6] = tmp1 + tmp2;
	(mat)[9] = tmp1 - tmp2;
}

LUX_INLINE const float *Matrix44GetIdentity()
{
	return _matrix44_ident;
}

LUX_INLINE void Matrix44Identity(Matrix44 m_mat)
{
	Matrix44Copy(m_mat,_matrix44_ident);
}

LUX_INLINE void Matrix44FromAngleAxis(Matrix44 matrix, float anglerad, const Vector3 axis)
{
	float c,s,oneminusc;
	c = fastcos(anglerad);
	s = fastsin(anglerad);
	oneminusc = 1.0f - c;
	Matrix44FromAngleAxisFast(matrix,axis,c,s,oneminusc);
}

LUX_INLINE void Matrix44PlaneProjection(Matrix44 matrix, const Vector3 planenormal)
{
	matrix[3] = matrix[7] = matrix[11] = 0.0f;
	matrix[15] = 1.0f;
	Matrix44SetRotRows(matrix,
		(planenormal[1]*planenormal[1])+(planenormal[2]*planenormal[2]),  	-planenormal[0]*planenormal[1],  	-planenormal[1]*planenormal[2],
		-planenormal[0]*planenormal[1],  	(planenormal[0]*planenormal[0])+(planenormal[2]*planenormal[2]),  	-planenormal[1]*planenormal[1],
		-planenormal[0]*planenormal[2],  	-planenormal[1]*planenormal[2],  	(planenormal[0]*planenormal[0])+(planenormal[1]*planenormal[1]));
}

LUX_INLINE void Matrix44VectorRotate( const Matrix44 mat, Vector3 v1 )
{
	float vector[3];

	vector[0] = v1[0]*mat[0]+v1[1]*mat[4]+v1[2]*mat[8];
	vector[1] = v1[0]*mat[1]+v1[1]*mat[5]+v1[2]*mat[9];
	vector[2] = v1[0]*mat[2]+v1[1]*mat[6]+v1[2]*mat[10];

	Vector3Copy(v1,vector);
}

LUX_INLINE void Matrix44VectorInvRotate( const Matrix44 m_mat, Vector3 pVect )
{
	float vec[3];

	vec[0] = pVect[0]*m_mat[0]+pVect[1]*m_mat[1]+pVect[2]*m_mat[2];
	vec[1] = pVect[0]*m_mat[4]+pVect[1]*m_mat[5]+pVect[2]*m_mat[6];
	vec[2] = pVect[0]*m_mat[8]+pVect[1]*m_mat[9]+pVect[2]*m_mat[10];

	Vector3Copy(pVect, vec);
}

LUX_INLINE void Matrix44VectorTranslate( const Matrix44 m_mat, Vector3 pVect )
{
	pVect[0] = pVect[0]+m_mat[12];
	pVect[1] = pVect[1]+m_mat[13];
	pVect[2] = pVect[2]+m_mat[14];
}

LUX_INLINE void Matrix44VectorInvTranslate( const Matrix44 m_mat, Vector3 pVect )
{
	pVect[0] = pVect[0]-m_mat[12];
	pVect[1] = pVect[1]-m_mat[13];
	pVect[2] = pVect[2]-m_mat[14];
}

LUX_INLINE void Matrix44VectorTransform( const Matrix44 mat, Vector3 v1 )
{
	Vector3Transform1(v1,mat);
}

LUX_INLINE void Matrix44VectorInvTransform( const Matrix44 mat, Vector3 v1 )
{
	Vector3InvTransform1(v1,mat);
}

LUX_INLINE float*	Matrix44GetTranslation(const Matrix44 mat, Vector3 vec){
	vec[0] = mat[12];
	vec[1] = mat[13];
	vec[2] = mat[14];
	return vec;
}

LUX_INLINE void Matrix44MultiplyRot(Matrix44 newmat, const float* LUX_RESTRICT m_mat, const float* LUX_RESTRICT mat )
{
	const float *a_mat = m_mat,*b_mat = mat;

	newmat[0] = a_mat[0]*b_mat[0] + a_mat[4]*b_mat[1] + a_mat[8]*b_mat[2];
	newmat[1] = a_mat[1]*b_mat[0] + a_mat[5]*b_mat[1] + a_mat[9]*b_mat[2];
	newmat[2] = a_mat[2]*b_mat[0] + a_mat[6]*b_mat[1] + a_mat[10]*b_mat[2];

	newmat[4] = a_mat[0]*b_mat[4] + a_mat[4]*b_mat[5] + a_mat[8]*b_mat[6];
	newmat[5] = a_mat[1]*b_mat[4] + a_mat[5]*b_mat[5] + a_mat[9]*b_mat[6];
	newmat[6] = a_mat[2]*b_mat[4] + a_mat[6]*b_mat[5] + a_mat[10]*b_mat[6];

	newmat[8] = a_mat[0]*b_mat[8] + a_mat[4]*b_mat[9] + a_mat[8]*b_mat[10];
	newmat[9] = a_mat[1]*b_mat[8] + a_mat[5]*b_mat[9] + a_mat[9]*b_mat[10];
	newmat[10] = a_mat[2]*b_mat[8] + a_mat[6]*b_mat[9] + a_mat[10]*b_mat[10];

}

LUX_INLINE void Matrix44Multiply( Matrix44 newmat, const float* LUX_RESTRICT m_mat,  const float* LUX_RESTRICT mat )
{
	const float *a_mat = m_mat,*b_mat = mat;

	newmat[0] = a_mat[0]*b_mat[0] + a_mat[4]*b_mat[1] + a_mat[8]*b_mat[2];
	newmat[1] = a_mat[1]*b_mat[0] + a_mat[5]*b_mat[1] + a_mat[9]*b_mat[2];
	newmat[2] = a_mat[2]*b_mat[0] + a_mat[6]*b_mat[1] + a_mat[10]*b_mat[2];
	newmat[3] = 0;

	newmat[4] = a_mat[0]*b_mat[4] + a_mat[4]*b_mat[5] + a_mat[8]*b_mat[6];
	newmat[5] = a_mat[1]*b_mat[4] + a_mat[5]*b_mat[5] + a_mat[9]*b_mat[6];
	newmat[6] = a_mat[2]*b_mat[4] + a_mat[6]*b_mat[5] + a_mat[10]*b_mat[6];
	newmat[7] = 0;

	newmat[8] = a_mat[0]*b_mat[8] + a_mat[4]*b_mat[9] + a_mat[8]*b_mat[10];
	newmat[9] = a_mat[1]*b_mat[8] + a_mat[5]*b_mat[9] + a_mat[9]*b_mat[10];
	newmat[10] = a_mat[2]*b_mat[8] + a_mat[6]*b_mat[9] + a_mat[10]*b_mat[10];
	newmat[11] = 0;

	newmat[12] = a_mat[0]*b_mat[12] + a_mat[4]*b_mat[13] + a_mat[8]*b_mat[14] + a_mat[12];
	newmat[13] = a_mat[1]*b_mat[12] + a_mat[5]*b_mat[13] + a_mat[9]*b_mat[14] + a_mat[13];
	newmat[14] = a_mat[2]*b_mat[12] + a_mat[6]*b_mat[13] + a_mat[10]*b_mat[14] + a_mat[14];
	newmat[15] = 1;
}

LUX_INLINE void Matrix44MultiplyFull( Matrix44 clip, const float* LUX_RESTRICT proj , const float* LUX_RESTRICT modl)
{

	clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
	clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
	clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
	clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];

	clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
	clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
	clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
	clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];

	clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
	clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
	clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
	clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];

	clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
	clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
	clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
	clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];

}

LUX_INLINE void Matrix44MultiplyRot1( float* LUX_RESTRICT m_mat, const float* LUX_RESTRICT mat )
{
	Matrix44 newmat;
	const float *a_mat,*b_mat;
	a_mat = m_mat; b_mat = mat;

	newmat[0] = a_mat[0]*b_mat[0] + a_mat[4]*b_mat[1] + a_mat[8]*b_mat[2];
	newmat[1] = a_mat[1]*b_mat[0] + a_mat[5]*b_mat[1] + a_mat[9]*b_mat[2];
	newmat[2] = a_mat[2]*b_mat[0] + a_mat[6]*b_mat[1] + a_mat[10]*b_mat[2];

	newmat[4] = a_mat[0]*b_mat[4] + a_mat[4]*b_mat[5] + a_mat[8]*b_mat[6];
	newmat[5] = a_mat[1]*b_mat[4] + a_mat[5]*b_mat[5] + a_mat[9]*b_mat[6];
	newmat[6] = a_mat[2]*b_mat[4] + a_mat[6]*b_mat[5] + a_mat[10]*b_mat[6];

	newmat[8] = a_mat[0]*b_mat[8] + a_mat[4]*b_mat[9] + a_mat[8]*b_mat[10];
	newmat[9] = a_mat[1]*b_mat[8] + a_mat[5]*b_mat[9] + a_mat[9]*b_mat[10];
	newmat[10] = a_mat[2]*b_mat[8] + a_mat[6]*b_mat[9] + a_mat[10]*b_mat[10];

	Matrix44CopyRot(m_mat,newmat);
}


LUX_INLINE void Matrix44Multiply1( float* LUX_RESTRICT m_mat,  const float* LUX_RESTRICT mat )
{
	Matrix44 newmat;
	const float *a_mat,*b_mat;
	a_mat = m_mat; b_mat = mat;

	newmat[0] = a_mat[0]*b_mat[0] + a_mat[4]*b_mat[1] + a_mat[8]*b_mat[2];
	newmat[1] = a_mat[1]*b_mat[0] + a_mat[5]*b_mat[1] + a_mat[9]*b_mat[2];
	newmat[2] = a_mat[2]*b_mat[0] + a_mat[6]*b_mat[1] + a_mat[10]*b_mat[2];
	newmat[3] = 0;

	newmat[4] = a_mat[0]*b_mat[4] + a_mat[4]*b_mat[5] + a_mat[8]*b_mat[6];
	newmat[5] = a_mat[1]*b_mat[4] + a_mat[5]*b_mat[5] + a_mat[9]*b_mat[6];
	newmat[6] = a_mat[2]*b_mat[4] + a_mat[6]*b_mat[5] + a_mat[10]*b_mat[6];
	newmat[7] = 0;

	newmat[8] = a_mat[0]*b_mat[8] + a_mat[4]*b_mat[9] + a_mat[8]*b_mat[10];
	newmat[9] = a_mat[1]*b_mat[8] + a_mat[5]*b_mat[9] + a_mat[9]*b_mat[10];
	newmat[10] = a_mat[2]*b_mat[8] + a_mat[6]*b_mat[9] + a_mat[10]*b_mat[10];
	newmat[11] = 0;

	newmat[12] = a_mat[0]*b_mat[12] + a_mat[4]*b_mat[13] + a_mat[8]*b_mat[14] + a_mat[12];
	newmat[13] = a_mat[1]*b_mat[12] + a_mat[5]*b_mat[13] + a_mat[9]*b_mat[14] + a_mat[13];
	newmat[14] = a_mat[2]*b_mat[12] + a_mat[6]*b_mat[13] + a_mat[10]*b_mat[14] + a_mat[14];
	newmat[15] = 1;

	Matrix44Copy(m_mat,newmat);
}

LUX_INLINE void Matrix44MultiplyRot2( const float* LUX_RESTRICT m_mat, float* LUX_RESTRICT mat )
{
	Matrix44 newmat;
	const float *a_mat,*b_mat;
	a_mat = m_mat; b_mat = mat;

	newmat[0] = a_mat[0]*b_mat[0] + a_mat[4]*b_mat[1] + a_mat[8]*b_mat[2];
	newmat[1] = a_mat[1]*b_mat[0] + a_mat[5]*b_mat[1] + a_mat[9]*b_mat[2];
	newmat[2] = a_mat[2]*b_mat[0] + a_mat[6]*b_mat[1] + a_mat[10]*b_mat[2];

	newmat[4] = a_mat[0]*b_mat[4] + a_mat[4]*b_mat[5] + a_mat[8]*b_mat[6];
	newmat[5] = a_mat[1]*b_mat[4] + a_mat[5]*b_mat[5] + a_mat[9]*b_mat[6];
	newmat[6] = a_mat[2]*b_mat[4] + a_mat[6]*b_mat[5] + a_mat[10]*b_mat[6];

	newmat[8] = a_mat[0]*b_mat[8] + a_mat[4]*b_mat[9] + a_mat[8]*b_mat[10];
	newmat[9] = a_mat[1]*b_mat[8] + a_mat[5]*b_mat[9] + a_mat[9]*b_mat[10];
	newmat[10] = a_mat[2]*b_mat[8] + a_mat[6]*b_mat[9] + a_mat[10]*b_mat[10];

	Matrix44CopyRot(mat,newmat);
}

LUX_INLINE void Matrix44Multiply2( const float* LUX_RESTRICT m_mat,  float*  LUX_RESTRICT mat )
{
	Matrix44 newmat;
	const float *a_mat,*b_mat;
	a_mat = m_mat; b_mat = mat;

	newmat[0] = a_mat[0]*b_mat[0] + a_mat[4]*b_mat[1] + a_mat[8]*b_mat[2];
	newmat[1] = a_mat[1]*b_mat[0] + a_mat[5]*b_mat[1] + a_mat[9]*b_mat[2];
	newmat[2] = a_mat[2]*b_mat[0] + a_mat[6]*b_mat[1] + a_mat[10]*b_mat[2];
	newmat[3] = 0;

	newmat[4] = a_mat[0]*b_mat[4] + a_mat[4]*b_mat[5] + a_mat[8]*b_mat[6];
	newmat[5] = a_mat[1]*b_mat[4] + a_mat[5]*b_mat[5] + a_mat[9]*b_mat[6];
	newmat[6] = a_mat[2]*b_mat[4] + a_mat[6]*b_mat[5] + a_mat[10]*b_mat[6];
	newmat[7] = 0;

	newmat[8] = a_mat[0]*b_mat[8] + a_mat[4]*b_mat[9] + a_mat[8]*b_mat[10];
	newmat[9] = a_mat[1]*b_mat[8] + a_mat[5]*b_mat[9] + a_mat[9]*b_mat[10];
	newmat[10] = a_mat[2]*b_mat[8] + a_mat[6]*b_mat[9] + a_mat[10]*b_mat[10];
	newmat[11] = 0;

	newmat[12] = a_mat[0]*b_mat[12] + a_mat[4]*b_mat[13] + a_mat[8]*b_mat[14] + a_mat[12];
	newmat[13] = a_mat[1]*b_mat[12] + a_mat[5]*b_mat[13] + a_mat[9]*b_mat[14] + a_mat[13];
	newmat[14] = a_mat[2]*b_mat[12] + a_mat[6]*b_mat[13] + a_mat[10]*b_mat[14] + a_mat[14];
	newmat[15] = 1;

	Matrix44Copy(mat,newmat);
}

LUX_INLINE void Matrix44SetTranslation( Matrix44 m_mat, const Vector3 translation )
{
	m_mat[12] = translation[0];
	m_mat[13] = translation[1];
	m_mat[14] = translation[2];
}

LUX_INLINE void Matrix44SetInvTranslation( Matrix44 m_mat, const Vector3 translation )
{
	m_mat[12] = -translation[0];
	m_mat[13] = -translation[1];
	m_mat[14] = -translation[2];
}

LUX_INLINE void Matrix44SetScale( Matrix44 mat, const Vector3 scale)
{
	mat[0] =scale[0];
	mat[5] =scale[1];
	mat[10] =scale[2];
}

LUX_INLINE void Matrix44Transpose1(Matrix44 mat)
{
	Matrix44 oldmat;
	Matrix44Copy(oldmat,mat);
	mat[1]=oldmat[4];
	mat[2]=oldmat[8];
	mat[3]=oldmat[12];
	mat[4]=oldmat[1];
	mat[8]=oldmat[2];
	mat[12]=oldmat[3];

	mat[6]=oldmat[9];
	mat[7]=oldmat[13];
	mat[9]=oldmat[6];
	mat[13]=oldmat[7];

	mat[11]=oldmat[14];
	mat[14]=oldmat[11];
}
LUX_INLINE void Matrix44TransposeRot1(Matrix44 mat)
{
	float tmp;
	tmp = mat[1]; mat[1] = mat[4]; mat[4] = tmp;
	tmp = mat[2]; mat[2] = mat[8]; mat[8] = tmp;
	tmp = mat[6]; mat[6] = mat[9]; mat[9] = tmp;
}

LUX_INLINE void Matrix44TransposeRot(Matrix44 mat, const float* LUX_RESTRICT oldmat)
{
	mat[0]=oldmat[0];
	mat[4]=oldmat[1];
	mat[8]=oldmat[2];

	mat[1]=oldmat[4];
	mat[5]=oldmat[5];
	mat[9]=oldmat[6];

	mat[2]=oldmat[8];
	mat[6]=oldmat[9];
	mat[10]=oldmat[10];
}

LUX_INLINE void Matrix44TransposeRotIdentity(Matrix44 mat, const float* LUX_RESTRICT oldmat)
{
	mat[0]=oldmat[0];
	mat[4]=oldmat[1];
	mat[8]=oldmat[2];
	mat[12]=0.0f;

	mat[1]=oldmat[4];
	mat[5]=oldmat[5];
	mat[9]=oldmat[6];
	mat[13]=0.0f;

	mat[2]=oldmat[8];
	mat[6]=oldmat[9];
	mat[10]=oldmat[10];
	mat[14]=0.0f;

	mat[3]=0.0f;
	mat[7]=0.0f;
	mat[11]=0.0f;
	mat[15]=1.0f;
}

LUX_INLINE void Matrix44Transpose(Matrix44 mat, const float* LUX_RESTRICT oldmat){
	mat[0]=oldmat[0];
	mat[4]=oldmat[1];
	mat[8]=oldmat[2];
	mat[12]=oldmat[3];

	mat[1]=oldmat[4];
	mat[5]=oldmat[5];
	mat[9]=oldmat[6];
	mat[13]=oldmat[7];

	mat[2]=oldmat[8];
	mat[6]=oldmat[9];
	mat[10]=oldmat[10];
	mat[14]=oldmat[11];

	mat[3]=oldmat[12];
	mat[7]=oldmat[13];
	mat[11]=oldmat[14];
	mat[15]=oldmat[15];
}


LUX_INLINE void Matrix44Invert(Matrix44 inverse, const float* LUX_RESTRICT src)
{
	int i, j, k, swap;
	float t;
	float temp[4][4];

	for (i=0; i<4; i++) {
		for (j=0; j<4; j++) {
			temp[i][j] = src[i*4+j];
		}
	}
	Matrix44Identity(inverse);

	for (i = 0; i < 4; i++) {
		/*
		** Look for largest element in column
		*/
		swap = i;
		for (j = i + 1; j < 4; j++) {
			if (fabs(temp[j][i]) > fabs(temp[i][i])) {
				swap = j;
			}
		}

		if (swap != i) {
			/*
			** Swap rows.
			*/
			for (k = 0; k < 4; k++) {
				t = temp[i][k];
				temp[i][k] = temp[swap][k];
				temp[swap][k] = t;

				t = inverse[i*4+k];
				inverse[i*4+k] = inverse[swap*4+k];
				inverse[swap*4+k] = t;
			}
		}

		if (temp[i][i] == 0) {
			/*
			** No non-zero pivot.  The matrix is singular, which shouldn't
			** happen.  This means the user gave us a bad matrix.
			*/
			return;
		}

		t = temp[i][i];
		for (k = 0; k < 4; k++) {
			temp[i][k] /= t;
			inverse[i*4+k] /= t;
		}
		for (j = 0; j < 4; j++) {
			if (j != i) {
				t = temp[j][i];
				for (k = 0; k < 4; k++) {
					temp[j][k] -= temp[i][k]*t;
					inverse[j*4+k] -= inverse[i*4+k]*t;
				}
			}
		}
	}
}

LUX_INLINE void Matrix44Invert_(Matrix44 inv_mat, const float* LUX_RESTRICT mat)
{
	float det;


	// calculate determinants of required 2x2 sub matrices
	float det01 =mat[9] *mat[14] -mat[13] *mat[10];		// these are required for the determinant
	float det02 =mat[5] *mat[14] -mat[13] *mat[6];
	float det03 =mat[5] *mat[10] -mat[9] *mat[6];
	float det04 =mat[1] *mat[14] -mat[13] *mat[2];
	float det05 =mat[1] *mat[10] -mat[9] *mat[2];
	float det06 =mat[1] *mat[6] -mat[5] *mat[2];

	det =  -mat[3] * (mat[4] * det01 -mat[8] * det02 +mat[12] * det03) +
		mat[7] * (mat[0] * det01 -mat[8] * det04 +mat[12] * det05) -
		mat[11] * (mat[0] * det02 -mat[4] * det04 +mat[12] * det06) +
		mat[15] * (mat[0] * det03 -mat[4] * det05 +mat[8] * det06);

	if (fabs(det) < FLOAT_EPSILON)
	{
		Matrix44Identity(inv_mat);
	}
	else
	{
		// calculate (transposed cofactor mat / determinant of given mat)
		float det07 =mat[10] *mat[15] -mat[14] *mat[11];		// these are additionally required determinants of 2x2 sub matrices for the cofactor mat
		float det08 =mat[14] *mat[7] -mat[6] *mat[15];
		float det09 =mat[6] *mat[11] -mat[10] *mat[7];
		float det10 =mat[8] *mat[15] -mat[12] *mat[11];
		float det11 =mat[12] *mat[7] -mat[4] *mat[15];
		float det12 =mat[4] *mat[11] -mat[8] *mat[7];
		float det13 =mat[8] *mat[13] -mat[12] *mat[9];
		float det14 =mat[12] *mat[5] -mat[4] *mat[13];
		float det15 =mat[4] *mat[9] -mat[8] *mat[5];
		float det16 =mat[2] *mat[15] -mat[14] *mat[3];
		float det17 =mat[10] *mat[3] -mat[2] *mat[11];
		float det18 =mat[0] *mat[15] -mat[12] *mat[3];
		float det19 =mat[8] *mat[3] -mat[0] *mat[11];
		float det20 =mat[0] *mat[13] -mat[12] *mat[1];
		float det21 =mat[8] *mat[1] -mat[0] *mat[9];
		float det22 =mat[2] *mat[7] -mat[6] *mat[3];
		float det23 =mat[0] *mat[7] -mat[4] *mat[3];
		float det24 =mat[0] *mat[5] -mat[4] *mat[1];

		det = 1 / det;

		inv_mat[0] = det * (mat[5] * det07 +mat[9] * det08 +mat[13] * det09);
		inv_mat[4] = det * (mat[6] * det10 +mat[10] * det11 +mat[14] * det12);
		inv_mat[8] = det * (mat[7] * det13 +mat[11] * det14 +mat[15] * det15);
		inv_mat[12] = det * (mat[4] * -det01 +mat[8] * det02 +mat[12] * -det03);

		inv_mat[1] = det * (mat[9] * det16 +mat[13] * det17 -mat[1] * det07);
		inv_mat[5] = det * (mat[10] * det18 +mat[14] * det19 -mat[2] * det10);
		inv_mat[9] = det * (mat[11] * det20 +mat[15] * det21 -mat[3] * det13);
		inv_mat[13] = det * (mat[8] * -det04 +mat[12] * det05 +mat[0] * det01);

		inv_mat[2] = det * (mat[13] * det22 -mat[1] * det08 -mat[5] * det16);
		inv_mat[6] = det * (mat[14] * det23 -mat[2] * det11 -mat[6] * det18);
		inv_mat[10] = det * (mat[15] * det24 -mat[3] * det14 -mat[7] * det20);
		inv_mat[14] = det * (mat[12] * -det06 +mat[0] * -det02 +mat[4] * det04);

		inv_mat[3] = det * (mat[1] * -det09 -mat[5] * det17 -mat[9] * det22);
		inv_mat[7] = det * (mat[2] * -det12 -mat[6] * det19 -mat[10] * det23);
		inv_mat[11] = det * (mat[3] * -det15 -mat[7] * det21 -mat[11] * det24);
		inv_mat[15] = det * (mat[0] * det03 +mat[4] * -det05 +mat[8] * det06);
	}

}

LUX_INLINE void Matrix44Invert1(Matrix44 m)
{
	Matrix44 invmat;
	Matrix44Invert(invmat,m);
	Matrix44Copy(m,invmat);
}

LUX_INLINE void Matrix44AffineInvert(Matrix44 outmat, const float* LUX_RESTRICT mat){
	//Matrix44 outmat;
	outmat[0] = mat[0];
	outmat[1] = mat[4];
	outmat[2] = mat[8];
	outmat[3] = 0.0f;
	outmat[4] = mat[1];
	outmat[5] = mat[5];
	outmat[6] = mat[9];
	outmat[7] = 0.0f;
	outmat[8] = mat[2];
	outmat[9] = mat[6];
	outmat[10] = mat[10];
	outmat[11] = 0.0f;
	outmat[12] = -(mat[0]*mat[12]+mat[1]*mat[13]+mat[2]*mat[14]);
	outmat[13] = -(mat[4]*mat[12]+mat[5]*mat[13]+mat[6]*mat[14]);
	outmat[14] = -(mat[8]*mat[12]+mat[9]*mat[13]+mat[10]*mat[14]);
	outmat[15] = 1.0f;
	//Matrix44Copy(mat,outmat);
}

LUX_INLINE void Matrix44RotateAroundVector(Matrix44 mat, Vector3 axis, float angleRad){
	float c = fastcos(angleRad);
	float s = fastsin(angleRad);
	float t = 1.0f - c;
	float u = axis[0];
	float v = axis[1];
	float w = axis[2];

	mat[0] =      c + u*u*t;
	mat[1] =  w * s + v*u*t;
	mat[2] = -v * s + w*u*t;
	mat[4] = -w * s + u*v*t;
	mat[5] =      c + v*v*t;
	mat[6] =  u * s + w*v*t;
	mat[8] =  v * s + u*w*t;
	mat[9] = -u * s + v*w*t;
	mat[10] =      c + w*w*t;

	/*
	mat[0] = c + axis[0]*axis[0]*t;
	mat[5] = c + axis[1]*axis[1]*t;
	mat[10] = c + axis[2]*axis[2]*t;

	tmp1 = axis[0]*axis[1]*t;
	tmp2 = axis[2]*s;
	mat[1] = tmp1 + tmp2;
	mat[4] = tmp1 - tmp2;

	tmp1 = axis[0]*axis[2]*t;
	tmp2 = axis[1]*s;
	mat[2] = tmp1 - tmp2;
	mat[8] = tmp1 + tmp2;

	tmp1 = axis[1]*axis[2]*t;
	tmp2 = axis[0]*s;
	mat[6] = tmp1 + tmp2;
	mat[9] = tmp1 - tmp2;
	*/
}




// ugly fixes to get things into OpenGL coords
LUX_INLINE void	Matrix443DSMAXtoGL(Matrix44 mat){
	Matrix44 a_mat;
	Matrix44Copy(a_mat,mat);

	// swap y,z
	// invert x

	mat[0] = -a_mat[0];
	mat[4] = -a_mat[4];
	mat[8] = -a_mat[8];
	mat[12]= -a_mat[12];
	mat[1] = a_mat[2];
	mat[5] = a_mat[6];
	mat[9] = a_mat[10];
	mat[13] = a_mat[14];
	mat[2]  = a_mat[1];
	mat[6]  = a_mat[5];
	mat[10] = a_mat[9];
	mat[14] = a_mat[13];
	mat[15] = 1;
	mat[3] = mat[7] = mat[11] = 0;
}
LUX_INLINE float  Matrix44CompareRot(Matrix44 a, Matrix44 b)
{
	float sum;
	int c,r;

	sum = 0;
	for (c = 0; c < 3; c++)
		for (r = 0; r < 3; r++){
			sum += (float)fabs((a[r+(c*4)] - b[r+(c*4)]));
		}

	return sum;
}

#ifdef LUX_COMPILER_MSC
#pragma warning( pop )
#endif

#ifdef __cplusplus
}
#endif

#endif

