// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt
//
// <Author: Christoph Kubisch>


#ifndef __LUXMATH_VECTOR2_H__
#define __LUXMATH_VECTOR2_H__

#include <luxmath/basetypes.h>
#include <luxmath/fastmath.h>

#ifdef __cplusplus
extern "C"{
#endif


void Vector2Set( Vector2 pV, float x, float y);
void Vector2Copy( Vector2 pOut, const Vector2 pV1);
void Vector2Clear( Vector2 pOut);
void Vector2Scale( Vector2 pOut, const Vector2 pV1, float val);

void Vector2Transform( Vector2 output, const Vector2 v1,  const Matrix44 mat );
void Vector2Transform1( Vector2 v1,  const Matrix44 mat );
void Vector2TransformRot( Vector2 output, const Vector2 v1,  const Matrix44 mat );
void Vector2TransformRot1( Vector2 output, const Matrix44 mat );
void Vector2InvTransform( Vector2 output, const Vector2 v1,  const Matrix44 mat ) ;
void Vector2InvTransform1( Vector2 v1,  const Matrix44 mat );

float Vector2LengthA( const Vector2 pV );
float Vector2Length( const Vector2 pV );
float Vector2SqLength( const Vector2 pV );
float Vector2Dot( const Vector2 pV1, const Vector2 pV2 );
float Vector2NormalizeA( Vector2 pOut, const Vector2 pV1 );
float Vector2Normalize( Vector2 pOut, const Vector2 pV1 );
float Vector2NormalizedA( Vector2 pOut );
float Vector2Normalized( Vector2 pOut );
//////////////////////////////////////////////////////////////////////////


LUX_INLINE void Vector2Set( Vector2 pV, float x, float y)
{
	pV[0] = x;
	pV[1] = y;
}
LUX_INLINE void Vector2Copy( Vector2 pOut, const Vector2 pV1)
{
	pOut[0] = pV1[0];
	pOut[1] = pV1[1];
}

LUX_INLINE void Vector2Clear( Vector2 pOut)
{
	pOut[0] = 0.0f;
	pOut[1] = 0.0f;
}

LUX_INLINE void Vector2Scale( Vector2 pOut, const Vector2 pV1, float val)
{
	pOut[0] = pV1[0] * val;
	pOut[1] = pV1[1] * val;
}

LUX_INLINE float Vector2LengthA( const Vector2 pV )
{
#ifdef __cplusplus
	return sqrtf(pV[0] * pV[0] + pV[1] * pV[1]);
#else
	return (float) sqrt(pV[0] * pV[0] + pV[1]);
#endif
}

LUX_INLINE float Vector2Length( const Vector2 pV )
{
	return fastsqrt(pV[0] * pV[0] + pV[1] * pV[1]);
}

LUX_INLINE float Vector2SqLength( const Vector2 pV )
{
	return pV[0] * pV[0] + pV[1] * pV[1] + pV[2];
}

LUX_INLINE float Vector2Dot( const Vector2 pV1, const Vector2 pV2 )
{

	return pV1[0] * pV2[0] + pV1[1] * pV2[1];
}

LUX_INLINE float Vector2NormalizeA( Vector2 pOut, const Vector2 pV1 )
{
	float length = Vector2LengthA(pV1);
	float lengthdiv = (length < 0.0001f) ? 1.0f : 1.0f/length;

	pOut[0] = pV1[0] * lengthdiv;
	pOut[1] = pV1[1] * lengthdiv;

	return length;
}

LUX_INLINE float Vector2Normalize( Vector2 pOut, const Vector2 pV1 )
{
	float lengthdiv = fastsqrtinv(Vector2SqLength(pV1));
	pOut[0] = pV1[0] * lengthdiv;
	pOut[1] = pV1[1] * lengthdiv;

	return 1.0f/lengthdiv;
}

LUX_INLINE float Vector2NormalizedA( Vector2 pOut )
{
	float length = Vector2LengthA(pOut);
	float lengthdiv = (length < 0.0001f) ? 1.0f : 1.0f/length;

	pOut[0] = pOut[0] * lengthdiv;
	pOut[1] = pOut[1] * lengthdiv;

	return length;
}

LUX_INLINE float Vector2Normalized( Vector2 pOut )
{
	float lengthdiv = fastsqrtinv(Vector2SqLength(pOut));
	pOut[0] = pOut[0] * lengthdiv;
	pOut[1] = pOut[1] * lengthdiv;

	return 1.0f/lengthdiv;
}

LUX_INLINE void Vector2Transform( Vector2 output, const Vector2 v1,  const Matrix44 mat ) 
{
	(output)[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[4]+(mat)[12];
	(output)[1] = (v1)[0]*(mat)[1]+(v1)[1]*(mat)[5]+(mat)[13];
}

LUX_INLINE void Vector2Transform1( Vector2 v1,  const Matrix44 mat ) 
{
	Vector2 output;

	(output)[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[4]+(mat)[12];
	(output)[1] = (v1)[0]*(mat)[1]+(v1)[1]*(mat)[5]+(mat)[13];
	Vector2Copy(v1,output);
}

LUX_INLINE void Vector2TransformRot( Vector2 output, const Vector2 v1,  const Matrix44 mat ) 
{
	(output)[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[4];
	(output)[1] = (v1)[0]*(mat)[1]+(v1)[1]*(mat)[5];
}

LUX_INLINE void Vector2TransformRot1( Vector2 v1,  const Matrix44 mat ) 
{
	Vector2 output;

	(output)[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[4];
	(output)[1] = (v1)[0]*(mat)[1]+(v1)[1]*(mat)[5];
	Vector2Copy(v1,output);
}

LUX_INLINE void Vector2InvTransform( Vector2 output, const Vector2 v1,  const Matrix44 mat ) 
{
	(output)[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[1]-(mat[0]*mat[12]+mat[1]*mat[13]+mat[2]*mat[14]);
	(output)[1] = (v1)[0]*(mat)[4]+(v1)[1]*(mat)[5]-(mat[4]*mat[12]+mat[5]*mat[13]+mat[6]*mat[14]);
}

LUX_INLINE void Vector2InvTransform1( Vector2 v1,  const Matrix44 mat ) 
{
	Vector2 output;

	(output)[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[4]+(mat)[12];
	(output)[1] = (v1)[0]*(mat)[1]+(v1)[1]*(mat)[5]+(mat)[13];
	Vector2Copy(v1,output);
}

#ifdef __cplusplus
}
#endif

#endif

