// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt
//
// <Author: Christoph Kubisch>


#ifndef __LUXMATH_VECTOR3_H__
#define __LUXMATH_VECTOR3_H__

#include <luxmath/basetypes.h>
#include <luxmath/fastmath.h>
#include <luxmath/misc.h>

#ifdef __cplusplus
extern "C"{
#endif

void Vector3Set( Vector3 pV, float x, float y, float z );
void Vector3Copy( Vector3 pOut, const Vector3 pV1);
void Vector3Clear( Vector3 pOut);

float Vector3Length( const Vector3 pV );
float Vector3LengthA( const Vector3 pV );
float Vector3SqLength( const Vector3 pV );
float Vector3Dot( const Vector3 pV1, const Vector3 pV2 );

void Vector3Add( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 );
void Vector3Sub( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 );
void Vector3Mul( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 );
void Vector3Div( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 );
void Vector3Min( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 );
void Vector3Max( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 );
void Vector3Rad2Deg(Vector3 dst, Vector3 src);
void Vector3Deg2Rad(Vector3 dst, Vector3 src);
void Vector3Random(Vector3 out);

void Vector3Cross( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 );
void Vector3Scale( Vector3 pOut, const Vector3 pV, float s);
void Vector3ScaledAdd( Vector3 pOut, const Vector3 pV, float s, const Vector3 pV2);
void Vector3MulAdd( Vector3 pOut, const Vector3 pV, const Vector3 pV2, const Vector3 pV3);
void Vector3Lerp( Vector3 pOut, const float t, const Vector3 pV1, const Vector3 pV2 );
void Vector3CatmullRom(Vector3 vout, const float t, const Vector3 v0, const Vector3 v1, const Vector3 v2, const Vector3 v3);
float Vector3Distance( const Vector3 a, const Vector3 b);
float Vector3DistanceA( const Vector3 a, const Vector3 b);
float Vector3SqDistance(const Vector3 a, const Vector3 b);

float Vector3GetRotateZcossin( const Vector3 in, float *cosf, float *sinf);
float Vector3GetRotateZcossinNormalized( const Vector3 in, float *cosf, float *sinf);
void  Vector3RotateZcossin( Vector3 out, const Vector3 in, float cosf, float sinf);

// fast but not precise, dont use in accumulated / matrix ops
float Vector3Normalized( Vector3 pOut );
float Vector3Normalize( Vector3 pOut, const Vector3 pV1 );

// more accurate version
float Vector3NormalizedA( Vector3 pOut );
float Vector3NormalizeA( Vector3 pOut, const Vector3 pV1 );

void Vector3Negate( Vector3 pOut, const Vector3 pV1);
void Vector3Negated( Vector3 pOut);
void Vector3Saturated( Vector3 pOut);

booln Vector3InsideAABB( const Vector3 point,  const Vector3 mins,  const Vector3 maxs);
booln Vector3IsZero(const Vector3 vec3);

void Vector3Transform( Vector3 output, const Vector3 v1,  const Matrix44 mat );
void Vector3Transform1( Vector3 v1,  const Matrix44 mat );
void Vector3InvTransform( Vector3 output, const Vector3 vector,  const Matrix44 mat );
void Vector3InvTransform1( Vector3 vector, const Matrix44 mat );
void Vector3TransformT( Vector3 output,  const Vector3 v1,  const Matrix44 mat );
void Vector3TransformT1( Vector3 v1, const Matrix44 mat );
void Vector3TransformRot( Vector3 output,  const Vector3 v1,  const Matrix44 mat );
void Vector3TransformRot1( Vector3 v1, const Matrix44 mat ) ;

float Vector3LineDistanceSq(const Vector3 point,const Vector3 linestart,const Vector3 linedir);
float Vector3LineDistanceSqFracc(const Vector3 point,
								 const Vector3 linestart, const Vector3 linedir, float *ofracc);
void Vector3Spread(Vector3 out, Vector3 in, const float radin, const float radout);
void Vector3Perpendicular(Vector3 out, Vector3 in);
void Vector3PerpendicularA(Vector3 out, Vector3 in);
void Vector3MinMax(Vector3 min, Vector3 max, Vector3 pos);

// normalized conversions: float is [-1/0,1] and others are [-TYPEMAX/0,TYPEMAX]
void Vector3float_FROM_ubyte( Vector3 vec3, const uchar ub3[3]);
void Vector3ubyte_FROM_float( uchar ub3[3], const Vector3 vec3);
void Vector3float_FROM_short( Vector3 vec3, const short shrt3[3]);
void Vector3short_FROM_float( short shrt3[3], const Vector3 vec3);
void Vector3float_FROM_ushort( Vector3 vec3, const ushort shrt3[3]);
void Vector3ushort_FROM_float( ushort shrt3[3], const Vector3 vec3);
void Vector3ubyte_FROM_ushort( uchar vec3[3], const ushort shrt3[3]);
void Vector3ushort_FROM_ubyte( ushort shrt3[3], const uchar vec3[3]);
void Vector3pkubyte_FROM_short(uchar pkub3[3], const short shrt3[3]);
void Vector3short_FROM_pkubyte(short shrt3[3], const uchar pkub3[3]);
void Vector3pkubyte_FROM_float(uchar pkub3[3], const Vector3 vec3);
void Vector3float_FROM_pkubyte(Vector3 vec3, const uchar pkub3[3]);

void Vector3Swizzle(Vector3 out, uint axis[3], Vector3 dirs);

//////////////////////////////////////////////////////////////////////////


LUX_INLINE void Vector3Set( Vector3 pV, float x, float y, float z )
{
	pV[0] = x;
	pV[1] = y;
	pV[2] = z;
}
LUX_INLINE void Vector3Copy( Vector3 pOut, const Vector3 pV1)
{
	pOut[0] = pV1[0];
	pOut[1] = pV1[1];
	pOut[2] = pV1[2];
}

LUX_INLINE void Vector3Clear( Vector3 pOut)
{
	pOut[0] = 0.0f;
	pOut[1] = 0.0f;
	pOut[2] = 0.0f;
}

LUX_INLINE float Vector3LengthA( const Vector3 pV )
{

#ifdef __cplusplus
	return sqrtf(pV[0] * pV[0] + pV[1] * pV[1] + pV[2] * pV[2]);
#else
	return (float) sqrt(pV[0] * pV[0] + pV[1] * pV[1] + pV[2] * pV[2]);
#endif
}

LUX_INLINE float Vector3Length( const Vector3 pV )
{

#ifdef __cplusplus
	return fastsqrt(pV[0] * pV[0] + pV[1] * pV[1] + pV[2] * pV[2]);
#else
	return (float) fastsqrt(pV[0] * pV[0] + pV[1] * pV[1] + pV[2] * pV[2]);
#endif
}

LUX_INLINE float Vector3SqLength( const Vector3 pV )
{
	return pV[0] * pV[0] + pV[1] * pV[1] + pV[2] * pV[2];
}

LUX_INLINE float Vector3Dot( const Vector3 pV1, const Vector3 pV2 )
{
	return pV1[0] * pV2[0] + pV1[1] * pV2[1] + pV1[2] * pV2[2];
}

LUX_INLINE void Vector3Cross( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 )
{
	pOut[0] = pV1[1] * pV2[2] - pV1[2] * pV2[1];
	pOut[1] = pV1[2] * pV2[0] - pV1[0] * pV2[2];
	pOut[2] = pV1[0] * pV2[1] - pV1[1] * pV2[0];
}

LUX_INLINE float Vector3NormalizeA( Vector3 pOut, const Vector3 pV1 )
{
	float length = Vector3LengthA(pV1);
	float lengthdiv = (length < 0.0001f) ? 1.0f : 1.0f/length;

	pOut[0] = pV1[0] * lengthdiv;
	pOut[1] = pV1[1] * lengthdiv;
	pOut[2] = pV1[2] * lengthdiv;

	return length;
}

LUX_INLINE float Vector3NormalizedA( Vector3 pOut )
{
	float length = Vector3LengthA(pOut);
	float lengthdiv = (length < 0.0001f) ? 1.0f : 1.0f/length;

	pOut[0] = pOut[0] * lengthdiv;
	pOut[1] = pOut[1] * lengthdiv;
	pOut[2] = pOut[2] * lengthdiv;

	return length;
}

LUX_INLINE float Vector3Normalize( Vector3 pOut, const Vector3 pV1 )
{
	//float length = 1.0f / fastsqrt(Vector3SqLength(pV1));
	float length = fastsqrtinv(Vector3SqLength(pV1));
	pOut[0] = pV1[0] * length;
	pOut[1] = pV1[1] * length;
	pOut[2] = pV1[2] * length;

	return 1.0f/length;
}

LUX_INLINE float Vector3Normalized( Vector3 pOut )
{
	//float length = 1.0f / fastsqrt(Vector3SqLength(pOut));
	float length = fastsqrtinv(Vector3SqLength(pOut));
	pOut[0] = pOut[0] * length;
	pOut[1] = pOut[1] * length;
	pOut[2] = pOut[2] * length;

	return 1.0f/length;
}


LUX_INLINE void Vector3Add( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 )
{
	pOut[0] = pV1[0] + pV2[0];
	pOut[1] = pV1[1] + pV2[1];
	pOut[2] = pV1[2] + pV2[2];
}

LUX_INLINE void Vector3Sub( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 )
{
	pOut[0] = pV1[0] - pV2[0];
	pOut[1] = pV1[1] - pV2[1];
	pOut[2] = pV1[2] - pV2[2];
}

LUX_INLINE void Vector3Mul( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 )
{
	pOut[0] = pV1[0] * pV2[0];
	pOut[1] = pV1[1] * pV2[1];
	pOut[2] = pV1[2] * pV2[2];
}

LUX_INLINE void Vector3Div( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 )
{
	pOut[0] = pV1[0] / pV2[0];
	pOut[1] = pV1[1] / pV2[1];
	pOut[2] = pV1[2] / pV2[2];
}

LUX_INLINE void Vector3Min( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 )
{
	pOut[0] = pV1[0] < pV2[0] ? pV1[0] : pV2[0];
	pOut[1] = pV1[1] < pV2[1] ? pV1[1] : pV2[1];
	pOut[2] = pV1[2] < pV2[2] ? pV1[2] : pV2[2];
}

LUX_INLINE void Vector3Max( Vector3 pOut, const Vector3 pV1, const Vector3 pV2 )
{
	pOut[0] = pV1[0] > pV2[0] ? pV1[0] : pV2[0];
	pOut[1] = pV1[1] > pV2[1] ? pV1[1] : pV2[1];
	pOut[2] = pV1[2] > pV2[2] ? pV1[2] : pV2[2];
}

LUX_INLINE void Vector3Scale( Vector3 pOut, const Vector3 pV, float s)
{
	pOut[0] = pV[0] * s;
	pOut[1] = pV[1] * s;
	pOut[2] = pV[2] * s;
}

LUX_INLINE void Vector3ScaledAdd( Vector3 pOut, const Vector3 pV, float s, const Vector3 pV2)
{
	pOut[0] = pV[0] + (s * pV2[0]);
	pOut[1] = pV[1] + (s * pV2[1]);
	pOut[2] = pV[2] + (s * pV2[2]);
}

LUX_INLINE void Vector3MulAdd( Vector3 pOut, const Vector3 pV, const Vector3 pV2, const Vector3 pV3)
{
	pOut[0] = pV[0] + (pV2[0] * pV3[0]);
	pOut[1] = pV[1] + (pV2[1] * pV3[1]);
	pOut[2] = pV[2] + (pV2[2] * pV3[2]);
}

LUX_INLINE void Vector3Lerp( Vector3 pOut, const float s, const Vector3 pV1, const Vector3 pV2 )
{
	pOut[0] = pV1[0] + s * (pV2[0] - pV1[0]);
	pOut[1] = pV1[1] + s * (pV2[1] - pV1[1]);
	pOut[2] = pV1[2] + s * (pV2[2] - pV1[2]);
}


LUX_INLINE void Vector3Negate( Vector3 pOut, const Vector3 pV1)
{
	pOut[0] = -pV1[0];
	pOut[1] = -pV1[1];
	pOut[2] = -pV1[2];
}

LUX_INLINE void Vector3Negated( Vector3 pOut)
{
	pOut[0] = -pOut[0];
	pOut[1] = -pOut[1];
	pOut[2] = -pOut[2];
}

LUX_INLINE void Vector3Saturated( Vector3 pOut)
{
	pOut[0] = (FP_GREATER_ONE(pOut[0])) ? 1.0f : (FP_LESS_ZERO(pOut[0])) ? 0.0f : pOut[0];
	pOut[1] = (FP_GREATER_ONE(pOut[1])) ? 1.0f : (FP_LESS_ZERO(pOut[1])) ? 0.0f : pOut[1];
	pOut[2] = (FP_GREATER_ONE(pOut[2])) ? 1.0f : (FP_LESS_ZERO(pOut[2])) ? 0.0f : pOut[2];
}

LUX_INLINE void Vector3Random(Vector3 out){
	booln zero;
	int i;

	zero = TRUE;
	while (zero){
		for (i = 0; i < 3; i++){
			out[i] = frand() - 0.5f;
			if (out[i] > FLOAT_EPSILON || out[i] < FLOAT_EPSILON)
				zero = FALSE;
		}
	}
	Vector3Normalized(out);
}

// returns 0
LUX_INLINE float Vector3GetRotateZcossinNormalized( const Vector3 in, float *cosf, float *sinf)
{
	float len;

	len = 0.0f;
	*cosf = in[0];
	*sinf = in[1];

	return len;
}

// returns length in X/Y
LUX_INLINE float Vector3GetRotateZcossin( const Vector3 in, float *cosf, float *sinf)
{
	Vector3 temp;
	float len;
	float out;

	Vector3Copy(temp,in);
	len = fastsqrt(temp[0]*temp[0]+temp[1]*temp[1]);
	out = 1/len;
	*cosf = temp[0]*out;
	*sinf = temp[1]*out;

	return len;
}

LUX_INLINE void Vector3RotateZcossin( Vector3 out, const Vector3 in, float cosf, float sinf)
{
	out[0] = cosf*in[0] - sinf*in[1];
	out[1] = sinf*in[0] + cosf*in[1];
	out[2] = in[2];
}

LUX_INLINE booln Vector3InsideAABB( const Vector3 point,  const Vector3 mins,  const Vector3 maxs)
{
	return (  !(point[0] < mins[0] ||	
				point[0] > maxs[0] ||	
				point[1] < mins[1] ||	
				point[1] > maxs[1] ||	
				point[2] < mins[2] ||	
				point[2] > maxs[2]) );
}

LUX_INLINE booln Vector3IsZero(const Vector3 vec)
{
	return (!(vec)[0] && !(vec)[1] && !(vec)[2]);
}

LUX_INLINE void Vector3CatmullRom(Vector3 vout, const float t, const Vector3 v0, const Vector3 v1, const Vector3 v2, const Vector3 v3)
{
	// catmull rom spline interpolation
	// between v1 and v2
	//q(t) = 0.5 *(    	(2 * P1) + (-P0 + P2) * t + (2*P0 - 5*P1 + 4*P2 - P3) * t2 + (-P0 + 3*P1- 3*P2 + P3) * t3)

	float t2=t*t;
	float t3=t2*t;

	float coeff0 = -t3 + 2.0f * t2 - t;
	float coeff1 = 3.0f * t3 - 5.0f * t2 + 2.0f;
	float coeff2 = -3.0f * t3 + 4.0f * t2 + t;
	float coeff3 = t3 - t2;

	vout[0] = (coeff0 * v0[0] + coeff1 * v1[0] + coeff2 * v2[0] + coeff3 * v3[0])*0.5f;
	vout[1] = (coeff0 * v0[1] + coeff1 * v1[1] + coeff2 * v2[1] + coeff3 * v3[1])*0.5f;
	vout[2] = (coeff0 * v0[2] + coeff1 * v1[2] + coeff2 * v2[2] + coeff3 * v3[2])*0.5f;
}


LUX_INLINE float Vector3Distance( const Vector3 a, const Vector3 b)
{
	Vector3	temp;

	Vector3Sub(temp,a,b);
	return Vector3Length(temp);
}

LUX_INLINE float Vector3DistanceA( const Vector3 a, const Vector3 b)
{
	Vector3	temp;

	Vector3Sub(temp,a,b);
	return Vector3LengthA(temp);
}

LUX_INLINE float Vector3SqDistance(const Vector3 a, const Vector3 b) 
{
	Vector3	temp;

	Vector3Sub(temp,a,b);
	return temp[0]*temp[0]+ temp[1]*temp[1]+ temp[2]*temp[2];
}
	

LUX_INLINE void Vector3Transform( Vector3 output, const Vector3 v1,  const Matrix44 mat ) 
{
	(output)[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[4]+(v1)[2]*(mat)[8]+(mat)[12];
	(output)[1] = (v1)[0]*(mat)[1]+(v1)[1]*(mat)[5]+(v1)[2]*(mat)[9]+(mat)[13];
	(output)[2] = (v1)[0]*(mat)[2]+(v1)[1]*(mat)[6]+(v1)[2]*(mat)[10]+(mat)[14];
}

LUX_INLINE void Vector3Transform1( Vector3 v1,  const Matrix44 mat ) 
{
	Vector3 vector;

	vector[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[4]+(v1)[2]*(mat)[8]+(mat)[12];
	vector[1] = (v1)[0]*(mat)[1]+(v1)[1]*(mat)[5]+(v1)[2]*(mat)[9]+(mat)[13];
	vector[2] = (v1)[0]*(mat)[2]+(v1)[1]*(mat)[6]+(v1)[2]*(mat)[10]+(mat)[14];
	Vector3Copy(v1,vector);
}

LUX_INLINE void Vector3InvTransform( Vector3 output, const Vector3 vector,  const Matrix44 mat ) 
{
	(output)[0] = (vector)[0]*(mat)[0]+(vector)[1]*(mat)[1]+(vector)[2]*(mat)[2]-(mat[0]*mat[12]+mat[1]*mat[13]+mat[2]*mat[14]);
	(output)[1] = (vector)[0]*(mat)[4]+(vector)[1]*(mat)[5]+(vector)[2]*(mat)[6]-(mat[4]*mat[12]+mat[5]*mat[13]+mat[6]*mat[14]);
	(output)[2] = (vector)[0]*(mat)[8]+(vector)[1]*(mat)[9]+(vector)[2]*(mat)[10]-(mat[8]*mat[12]+mat[9]*mat[13]+mat[10]*mat[14]);
}

LUX_INLINE void Vector3InvTransform1( Vector3 v1,  const Matrix44 mat ) 
{
	Vector3 vector;

	vector[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[1]+(v1)[2]*(mat)[2]-(mat[0]*mat[12]+mat[1]*mat[13]+mat[2]*mat[14]);
	vector[1] = (v1)[0]*(mat)[4]+(v1)[1]*(mat)[5]+(v1)[2]*(mat)[6]-(mat[4]*mat[12]+mat[5]*mat[13]+mat[6]*mat[14]);
	vector[2] = (v1)[0]*(mat)[8]+(v1)[1]*(mat)[9]+(v1)[2]*(mat)[10]-(mat[8]*mat[12]+mat[9]*mat[13]+mat[10]*mat[14]);
	Vector3Copy(v1,vector);
}

LUX_INLINE void Vector3TransformT( Vector3 output,  const Vector3 v1,  const Matrix44 mat )
{
	(output)[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[1]+(v1)[2]*(mat)[2]+(mat)[3];
	(output)[1] = (v1)[0]*(mat)[4]+(v1)[1]*(mat)[5]+(v1)[2]*(mat)[6]+(mat)[7];
	(output)[2] = (v1)[0]*(mat)[8]+(v1)[1]*(mat)[9]+(v1)[2]*(mat)[10]+(mat)[11];
}

LUX_INLINE void Vector3TransformT1( Vector3 v1, const Matrix44 mat ) 
{
	Vector3 vector;

	vector[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[1]+(v1)[2]*(mat)[2]+(mat)[3];
	vector[1] = (v1)[0]*(mat)[4]+(v1)[1]*(mat)[5]+(v1)[2]*(mat)[6]+(mat)[7];
	vector[2] = (v1)[0]*(mat)[8]+(v1)[1]*(mat)[9]+(v1)[2]*(mat)[10]+(mat)[11];
	Vector3Copy(v1,vector);
}

LUX_INLINE void Vector3TransformRot( Vector3 output,  const Vector3 v1,  const Matrix44 mat )
{
	(output)[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[4]+(v1)[2]*(mat)[8];
	(output)[1] = (v1)[0]*(mat)[1]+(v1)[1]*(mat)[5]+(v1)[2]*(mat)[9];
	(output)[2] = (v1)[0]*(mat)[2]+(v1)[1]*(mat)[6]+(v1)[2]*(mat)[10];
}

LUX_INLINE void Vector3TransformRot1( Vector3 v1, const Matrix44 mat ) 
{
	Vector3 vector;

	vector[0] = (v1)[0]*(mat)[0]+(v1)[1]*(mat)[4]+(v1)[2]*(mat)[8];
	vector[1] = (v1)[0]*(mat)[1]+(v1)[1]*(mat)[5]+(v1)[2]*(mat)[9];
	vector[2] = (v1)[0]*(mat)[2]+(v1)[1]*(mat)[6]+(v1)[2]*(mat)[10];
	Vector3Copy(v1,vector);
}


LUX_INLINE void Vector3Rad2Deg(Vector3 dst, Vector3 src)
{
	(dst)[0]=RAD2DEG((src)[0]);
	(dst)[1]=RAD2DEG((src)[1]);
	(dst)[2]=RAD2DEG((src)[2]);
}

LUX_INLINE void Vector3Deg2Rad(Vector3 dst, Vector3 src)
{
	(dst)[0]=DEG2RAD((src)[0]);
	(dst)[1]=DEG2RAD((src)[1]);
	(dst)[2]=DEG2RAD((src)[2]);
}

LUX_INLINE void Vector3float_FROM_ubyte( Vector3 vec3, const uchar ub3[3]) 
{
	(vec3)[0] = (float)(ub3)[0]*DIV_255;
	(vec3)[1] = (float)(ub3)[1]*DIV_255;
	(vec3)[2] = (float)(ub3)[2]*DIV_255;
}

LUX_INLINE void Vector3ubyte_FROM_float( uchar ub3[3], const Vector3 vec3) 
{
	FP_NORM_TO_BYTE((ub3)[0],(vec3)[0]);
	FP_NORM_TO_BYTE((ub3)[1],(vec3)[1]);
	FP_NORM_TO_BYTE((ub3)[2],(vec3)[2]);
}

LUX_INLINE void Vector3float_FROM_short( Vector3 vec3, const short shrt3[3]) 
{
	(vec3)[0]= DIV_SHORT*(float)(shrt3)[0];
	(vec3)[1]= DIV_SHORT*(float)(shrt3)[1];
	(vec3)[2]= DIV_SHORT*(float)(shrt3)[2];
}

LUX_INLINE void Vector3short_FROM_float( short shrt3[3], const Vector3 vec3) 
{
	(shrt3)[0]= (short)(M_SHORT*(vec3)[0]);
	(shrt3)[1]= (short)(M_SHORT*(vec3)[1]);
	(shrt3)[2]= (short)(M_SHORT*(vec3)[2]);
}

LUX_INLINE void Vector3float_FROM_ushort( Vector3 vec3, const ushort shrt3[3]) 
{
	(vec3)[0]= DIV_USHORT*(float)(shrt3)[0];
	(vec3)[1]= DIV_USHORT*(float)(shrt3)[1];
	(vec3)[2]= DIV_USHORT*(float)(shrt3)[2];
}

LUX_INLINE void Vector3ushort_FROM_float( ushort shrt3[3], const Vector3 vec3) 
{
	(shrt3)[0]= (short)(M_USHORT*(vec3)[0]);
	(shrt3)[1]= (short)(M_USHORT*(vec3)[1]);
	(shrt3)[2]= (short)(M_USHORT*(vec3)[2]);
}

LUX_INLINE void Vector3ubyte_FROM_ushort( uchar vec3[3], const ushort shrt3[3]) 
{
	(vec3)[0]= (shrt3)[0]>>8;
	(vec3)[1]= (shrt3)[1]>>8;
	(vec3)[2]= (shrt3)[2]>>8;
}

LUX_INLINE void Vector3ushort_FROM_ubyte( ushort shrt3[3], const uchar vec3[3]) 
{
	(shrt3)[0]= (vec3)[0]<<8;
	(shrt3)[1]= (vec3)[1]<<8;
	(shrt3)[2]= (vec3)[2]<<8;
}

LUX_INLINE void Vector3pkubyte_FROM_short(uchar pkub3[3], const short shrt3[3]){
	pkub3[0] = ((((int)shrt3[0])/2)+SHORT_SIGNEDMAX)>>8;
	pkub3[1] = ((((int)shrt3[1])/2)+SHORT_SIGNEDMAX)>>8;
	pkub3[2] = ((((int)shrt3[2])/2)+SHORT_SIGNEDMAX)>>8;
}

LUX_INLINE void Vector3short_FROM_pkubyte(short shrt3[3], const uchar pkub3[3]){
	shrt3[0] = (((short)pkub3[0])-128)*256;
	shrt3[1] = (((short)pkub3[1])-128)*256;
	shrt3[2] = (((short)pkub3[2])-128)*256;
}

LUX_INLINE void Vector3pkubyte_FROM_float(uchar pkub3[3], const Vector3 vec3){
	Vector3 temp;
	Vector3 offset = {0.5f,0.5f,0.5f};
	Vector3ScaledAdd(temp,offset,0.5f,vec3);
	Vector3ubyte_FROM_float(pkub3,temp);
}

LUX_INLINE void Vector3float_FROM_pkubyte(Vector3 vec3, const uchar pkub3[3]){
	Vector3 temp;
	Vector3 offset = {-1.0f,-1.0f,-1.0f};
	Vector3float_FROM_ubyte(temp,pkub3);
	Vector3ScaledAdd(vec3,offset,2.0f,temp);
}

LUX_INLINE void Vector3Swizzle(Vector3 out, uint axis[3], Vector3 dirs)
{
	Vector3 vector;
	Vector3Copy(vector,out);
	out[0] = dirs[0]*vector[axis[0]];
	out[1] = dirs[1]*vector[axis[1]];
	out[2] = dirs[2]*vector[axis[2]];
}

#ifdef __cplusplus
}
#endif

#endif
