// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt
//
// <Author: Christoph Kubisch>
// Some SIMD code taken from "Nebula2SDK", (C) 2002 RadonLabs GmbH

#ifndef __LUXMATH_XMM_H__
#define __LUXMATH_XMM_H__

#include <luxmath/basetypes.h>

#ifdef __cplusplus
extern "C"{
#endif

// Type_func instead of TypeFunc because Type is not native, (struct/aligned mem...)
// some functions were speed-test against normal fpu on a intel quadcore, vc2005

#ifdef LUX_SIMD_SSE
typedef __m128* Vector3SSE;
typedef __m128* Vector4SSE;
typedef __m128* Matrix44SSE;
typedef __m128* Matrix34SSE;

//////////////////////////////////////////////////////////////////////////
// Vector3SSE
// reads all 16-byte ( so make sure memory acces is allowed)
// but protects .w on write, 

void Vector3SSE_transform(Vector3SSE out,  Vector3SSE src, Matrix44SSE matrix);		// 1.7x
void Vector3SSE_transformRot(Vector3SSE out,  Vector3SSE src, Matrix44SSE matrix);	// 1.7x
void Vector3SSE_lerp(Vector3SSE out,  float s, Vector3SSE a, Vector3SSE b);
void Vector3SSE_catmullRom(Vector3SSE out,  float s, Vector3SSE prev, Vector3SSE a,Vector3SSE b, Vector3SSE next);

//////////////////////////////////////////////////////////////////////////
// Vector4SSE
void Vector4SSE_clear(Vector4SSE out);
void Vector4SSE_scale(Vector4SSE out,  float s);
void Vector4SSE_add(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1);
void Vector4SSE_sub(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1);
void Vector4SSE_mul(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1);
void Vector4SSE_div(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1);
void Vector4SSE_min(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1);
void Vector4SSE_max(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1);
void Vector4SSE_negated(Vector4SSE out);
void Vector4SSE_transform(Vector4SSE out,  Vector4SSE src, Matrix44SSE matrix);
void Vector4SSE_transformRot(Vector4SSE out,  Vector4SSE src, Matrix44SSE matrix);
void Vector4SSE_lerp(Vector4SSE out,  float s, Vector4SSE a, Vector4SSE b);
void Vector4SSE_catmullRom(Vector4SSE out,  float s, Vector4SSE prev, Vector4SSE a,Vector4SSE b, Vector4SSE next);

// ignores .w
// but may write to all
void  Vector4SSE_transformV3(Vector4SSE vec3,  Vector4SSE src, Matrix44SSE matrix);	// 1.8x
void  Vector4SSE_crossV3(Vector4SSE vec3,  Vector4SSE v0,  Vector4SSE v1);
void  Vector4SSE_normalizeV3(Vector4SSE out,  Vector4SSE vec3);
void  Vector4SSE_normalizedV3(Vector4SSE vec3);
float Vector4SSE_lengthV3( Vector4SSE vec3);
float Vector4SSE_sqLengthV3( Vector4SSE vec3);


//////////////////////////////////////////////////////////////////////////
// Matrix44SSE
void Matrix44SSE_clear(Matrix44SSE matrix);
void Matrix44SSE_identity(Matrix44SSE matrix);							// 2.0x
void Matrix44SSE_copy(Matrix44SSE matrix,  Matrix44SSE from);			// 1.9x
void Matrix44SSE_transpose1(Matrix44SSE matrix);						// 2.5x
void Matrix44SSE_transpose(Matrix44SSE matrix,Matrix44SSE from);
void Matrix44SSE_rotateVector(Matrix44SSE matrix, Vector4SSE src);
void Matrix44SSE_affineInvert(Matrix44SSE outmat,  Matrix44SSE mat);	// 1.6x

void Matrix44SSE_multiply(Matrix44SSE output,  Matrix44SSE mat0,  Matrix44SSE mat1); // 2.5x vs simple, 3.9x vs full
void Matrix44SSE_multiply1(Matrix44SSE outmat0,  Matrix44SSE mat1);
void Matrix44SSE_multiply2(Matrix44SSE mat0, Matrix44SSE outmat1);
void Matrix44SSE_invert(Matrix44SSE outmat, Matrix44SSE mat0);

//////////////////////////////////////////////////////////////////////////
// Matrix34SSE
void Matrix34SSE_identity(Matrix34SSE matrix);
void Matrix34SSE_copy(Matrix34SSE dst, Matrix34SSE src);
void Matrix34SSE_tmultiply44(Matrix34SSE output, Matrix44SSE mat0,  Matrix44SSE mat1);

//////////////////////////////////////////////////////////////////////////
// Some SIMD code taken from "Nebula2SDK"
//	(C) 2002 RadonLabs GmbH

//////////////////////////////////////////////////////////////////////////
// Vector4SSE


LUX_INLINE void Vector4SSE_clear(Vector4SSE out)
{
	*out = _mm_setzero_ps();
}

LUX_INLINE void Vector4SSE_scale(Vector4SSE out, float s)
{
	__m128 packed = _mm_set1_ps(s);
	*out=_mm_mul_ps(*out, packed);
}

LUX_INLINE void Vector4SSE_add(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1)
{
	*out = _mm_add_ps(*v0, *v1);
}

LUX_INLINE void Vector4SSE_sub(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1)
{
	*out=_mm_sub_ps(*v0, *v1);
}

LUX_INLINE void Vector4SSE_mul(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1)
{
	*out = _mm_mul_ps(*v0, *v1);
}

LUX_INLINE void Vector4SSE_div(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1)
{
	*out = _mm_div_ps(*v0, *v1);
}

LUX_INLINE void Vector4SSE_min(Vector4SSE out, Vector4SSE v0, Vector4SSE v1)
{
	*out = _mm_min_ps(*v0, *v1);
}

LUX_INLINE void Vector4SSE_max(Vector4SSE out, Vector4SSE v0, Vector4SSE v1)
{
	*out = _mm_max_ps(*v0, *v1);
}

LUX_INLINE void Vector4SSE_negated(Vector4SSE out)
{
	__m128 zero = _mm_setzero_ps();
	*out=_mm_sub_ps(zero, *out);
}

LUX_INLINE void Vector4SSE_transformRot(Vector4SSE dst, Vector4SSE src, Matrix44SSE matrix)
{
	*dst = _mm_add_ps(
		_mm_add_ps(
		_mm_mul_ps(_mm_shuffle_ps(*src, *src, _MM_SHUFFLE(0,0,0,0)), matrix[0]), 
		_mm_mul_ps(_mm_shuffle_ps(*src, *src, _MM_SHUFFLE(1,1,1,1)), matrix[1])), 
		_mm_mul_ps(_mm_shuffle_ps(*src, *src, _MM_SHUFFLE(2,2,2,2)), matrix[2]));
}

LUX_INLINE void Vector4SSE_transform(Vector4SSE dst,  Vector4SSE src, Matrix44SSE matrix)
{
	*dst = _mm_add_ps(
		_mm_add_ps(
		_mm_add_ps(
		_mm_mul_ps(_mm_shuffle_ps(*src, *src, _MM_SHUFFLE(0,0,0,0)), matrix[0]), 
		_mm_mul_ps(_mm_shuffle_ps(*src, *src, _MM_SHUFFLE(1,1,1,1)), matrix[1])), 
		_mm_mul_ps(_mm_shuffle_ps(*src, *src, _MM_SHUFFLE(2,2,2,2)), matrix[2])),
		_mm_mul_ps(_mm_shuffle_ps(*src, *src, _MM_SHUFFLE(3,3,3,3)), matrix[3]));
}

LUX_INLINE void Vector4SSE_lerp(Vector4SSE dst,  float s, Vector4SSE a, Vector4SSE b)
{
	// a + (b-a)*s

	*dst = _mm_add_ps(*a,_mm_mul_ps(_mm_sub_ps(*b,*a),_mm_set_ps1(s)));
}

LUX_INLINE void Vector4SSE_catmullRom(Vector4SSE dst,  float t, Vector4SSE v0, Vector4SSE v1,Vector4SSE v2, Vector4SSE v3)
{
	//  0.5 *(    	(2 * P1) + (P2 - P0) * t + (2*P0 - 5*P1 + 4*P2 - P3) * t2 + (3*P1-3*P2 + P3 -P0) * t3)

	float t2=t*t;
	float t3=t2*t;

	float coeff0 = -t3 + 2.0f * t2 - t;
	float coeff1 = 3.0f * t3 - 5.0f * t2 + 2.0f;
	float coeff2 = -3.0f * t3 + 4.0f * t2 + t;
	float coeff3 = t3 - t2;

	*dst = _mm_mul_ps(
			_mm_add_ps(
				_mm_mul_ps(*v0,_mm_set_ps1(coeff0)),
				_mm_add_ps(
					_mm_mul_ps(*v1,_mm_set_ps1(coeff1)),
					_mm_add_ps(
						_mm_mul_ps(*v2,_mm_set_ps1(coeff2)),
						_mm_mul_ps(*v3,_mm_set_ps1(coeff3))))),
			_mm_set_ps1(0.5f));
}


// ignore .w
LUX_INLINE void  Vector4SSE_transformV3(Vector4SSE vec3,  Vector4SSE src, Matrix44SSE matrix)
{
	*vec3 = _mm_add_ps(
		_mm_add_ps(
		_mm_add_ps(
		_mm_mul_ps(_mm_shuffle_ps(*src, *src, _MM_SHUFFLE(0,0,0,0)), matrix[0]), 
		_mm_mul_ps(_mm_shuffle_ps(*src, *src, _MM_SHUFFLE(1,1,1,1)), matrix[1])), 
		_mm_mul_ps(_mm_shuffle_ps(*src, *src, _MM_SHUFFLE(2,2,2,2)), matrix[2])),
		matrix[3]);
}
LUX_INLINE void Vector4SSE_crossV3(Vector4SSE out,  Vector4SSE v0,  Vector4SSE v1)
{
	__m128 a = _mm_shuffle_ps(*v0, *v0, _MM_SHUFFLE(3, 0, 2, 1));
	__m128 b = _mm_shuffle_ps(*v1, *v1, _MM_SHUFFLE(3, 1, 0, 2));
	__m128 c = _mm_shuffle_ps(*v0, *v0, _MM_SHUFFLE(3, 1, 0, 2));
	__m128 d = _mm_shuffle_ps(*v1, *v1, _MM_SHUFFLE(3, 0, 2, 1));

	__m128 e = _mm_mul_ps(a, b);
	__m128 f = _mm_mul_ps(c, d);

	*out = _mm_sub_ps(e, f);
}

LUX_INLINE void Vector4SSE_normalizeV3(Vector4SSE out, Vector4SSE vec3)
{
	__m128 a = _mm_mul_ps(*vec3, *vec3);

	// horizontal add
	__m128 b = _mm_add_ss(_mm_shuffle_ps(a, a, _MM_SHUFFLE(0,0,0,0)), _mm_add_ss(_mm_shuffle_ps(a, a, _MM_SHUFFLE(1,1,1,1)), _mm_shuffle_ps(a, a, _MM_SHUFFLE(2,2,2,2))));

	// get reciprocal of square root of squared length
	__m128 f = _mm_rsqrt_ss(b);
	__m128 oneDivLen = _mm_shuffle_ps(f, f, _MM_SHUFFLE(0,0,0,0));

	*out = _mm_mul_ps(*vec3, oneDivLen);
}

LUX_INLINE void Vector4SSE_normalizedV3(Vector4SSE vec3)
{
	__m128 a = _mm_mul_ps(*vec3, *vec3);

	// horizontal add
	__m128 b = _mm_add_ss(_mm_shuffle_ps(a, a, _MM_SHUFFLE(0,0,0,0)), _mm_add_ss(_mm_shuffle_ps(a, a, _MM_SHUFFLE(1,1,1,1)), _mm_shuffle_ps(a, a, _MM_SHUFFLE(2,2,2,2))));

	// get reciprocal of square root of squared length
	__m128 f = _mm_rsqrt_ss(b);
	__m128 oneDivLen = _mm_shuffle_ps(f, f, _MM_SHUFFLE(0,0,0,0));

	*vec3 = _mm_mul_ps(*vec3, oneDivLen);
}

LUX_INLINE float Vector4SSE_lengthV3(Vector4SSE vec3)
{
	__m128 a = _mm_mul_ps(*vec3, *vec3);

	// horizontal add
	__m128 b = _mm_add_ss(_mm_shuffle_ps(a, a, _MM_SHUFFLE(0,0,0,0)), _mm_add_ss(_mm_shuffle_ps(a, a, _MM_SHUFFLE(1,1,1,1)), _mm_shuffle_ps(a, a, _MM_SHUFFLE(2,2,2,2))));
	__m128 l = _mm_sqrt_ss(b);
	return l.m128_f32[0];
}

LUX_INLINE float Vector4SSE_sqLengthV3( Vector4SSE  vec3)
{
	__m128 a = _mm_mul_ps(*vec3, *vec3);
	__m128 b = _mm_add_ss(_mm_shuffle_ps(a, a, _MM_SHUFFLE(0,0,0,0)), _mm_add_ss(_mm_shuffle_ps(a, a, _MM_SHUFFLE(1,1,1,1)), _mm_shuffle_ps(a, a, _MM_SHUFFLE(2,2,2,2))));
	return b.m128_f32[0];
}

//////////////////////////////////////////////////////////////////////////
// Vector3SSE

//#define SSE_STORE3_SIMD
LUX_INLINE void Vector3SSE_transform(Vector3SSE out,  Vector3SSE src, Matrix44SSE matrix)
{
#ifndef SSE_STORE3_SIMD
	float old = out->m128_f32[3];
	Vector4SSE_transformV3(out,src,matrix);
	out->m128_f32[3] = old;
#else
	__m128 temp;
	Vector4SSE_transformV3(temp,src,matrix);

	temp = _mm_shuffle_ps(temp,temp,_MM_SHUFFLE(2,3,0,1));
	_mm_store_ss(((float*)out)+2,temp);
	_mm_storeh_pi((__m64*)out,temp);
#endif
}
#undef SSE_STORE3_SIMD

LUX_INLINE void Vector3SSE_transformRot(Vector3SSE out,  Vector3SSE src, Matrix44SSE matrix)
{
	float old = out->m128_f32[3];
	Vector4SSE_transformRot(out,src,matrix);
	out->m128_f32[3] = old;
}
LUX_INLINE void Vector3SSE_lerp(Vector3SSE out,  float s, Vector3SSE a, Vector3SSE b)
{
	float old = out->m128_f32[3];
	Vector4SSE_lerp(out,s,a,b);
	out->m128_f32[3] = old;
}
LUX_INLINE void Vector3SSE_catmullRom(Vector3SSE out,  float s, Vector3SSE v0, Vector3SSE v1,Vector3SSE v2, Vector3SSE v3)
{
	float old = out->m128_f32[3];
	Vector4SSE_catmullRom(out,s,v0,v1,v2,v3);
	out->m128_f32[3] = old;
}

// not worth the penalty
LUX_INLINE void Vector3SSE_add(Vector3SSE out,  Vector3SSE src, Vector3SSE src0)
{
	float old = out->m128_f32[3];
	*out  = _mm_add_ps(*src,*src0);
	out->m128_f32[3] = old;
}

LUX_INLINE void Vector3SSE_mul(Vector3SSE out,  Vector3SSE src, Vector3SSE src0)
{
	float old = out->m128_f32[3];
	*out  = _mm_mul_ps(*src,*src0);
	out->m128_f32[3] = old;
}


//////////////////////////////////////////////////////////////////////////
// Matrix

LUX_INLINE void Matrix44SSE_clear(Matrix44SSE matrix)
{
	matrix[0]=_mm_setzero_ps();
	matrix[1]=_mm_setzero_ps();
	matrix[2]=_mm_setzero_ps();
	matrix[3]=_mm_setzero_ps();
}

LUX_INLINE void Matrix44SSE_identity(Matrix44SSE matrix)
{
	matrix[0]=_mm_set_ps(0.0f,0.0f,0.0f,1.0f);
	matrix[1]=_mm_set_ps(0.0f,0.0f,1.0f,0.0f);
	matrix[2]=_mm_set_ps(0.0f,1.0f,0.0f,0.0f);
	matrix[3]=_mm_set_ps(1.0f,0.0f,0.0f,0.0f);
	/*
	matrix[0]=_mm_setzero_ps();
	matrix[1]=_mm_setzero_ps();
	matrix[2]=_mm_setzero_ps();
	matrix[3]=_mm_setzero_ps();

	matrix[0].m128_f32[0] = 
		matrix[1].m128_f32[1] = 
		matrix[2].m128_f32[2] = 
		matrix[3].m128_f32[3] = 1.0f;
	*/
}

LUX_INLINE void Matrix44SSE_copy(Matrix44SSE matrix, Matrix44SSE from)
{
	
	matrix[0] = from[0];
	matrix[1] = from[1];
	matrix[2] = from[2];
	matrix[3] = from[3];
}

LUX_INLINE void Matrix44SSE_transpose1(Matrix44SSE matrix)
{
	_MM_TRANSPOSE4_PS(matrix[0], matrix[1], matrix[2], matrix[3]);
}

LUX_INLINE void Matrix44SSE_transpose(Matrix44SSE matrix,Matrix44SSE from)
{
	__m128 tmp3, tmp2, tmp1, tmp0;                         
	tmp0   = _mm_shuffle_ps(from[0], from[1], 0x44); 
	tmp2   = _mm_shuffle_ps(from[0], from[1], 0xEE);
	tmp1   = _mm_shuffle_ps(from[2], from[3], 0x44);
	tmp3   = _mm_shuffle_ps(from[2], from[3], 0xEE); 

	matrix[0] = _mm_shuffle_ps(tmp0, tmp1, 0x88);
	matrix[1] = _mm_shuffle_ps(tmp0, tmp1, 0xDD);
	matrix[2] = _mm_shuffle_ps(tmp2, tmp3, 0x88);
	matrix[3] = _mm_shuffle_ps(tmp2, tmp3, 0xDD);
}

LUX_INLINE void Matrix44SSE_rotateVector(Matrix44SSE matrix, Vector4SSE src)
{
	__m128 m128 = *src;
	*src =	_mm_add_ps(
		_mm_add_ps(
		_mm_mul_ps(_mm_shuffle_ps(m128, m128, _MM_SHUFFLE(0,0,0,0)), matrix[0]), 
		_mm_mul_ps(_mm_shuffle_ps(m128, m128, _MM_SHUFFLE(1,1,1,1)), matrix[1])), 
		_mm_mul_ps(_mm_shuffle_ps(m128, m128, _MM_SHUFFLE(2,2,2,2)), matrix[2]));

}


LUX_INLINE void Matrix44SSE_multiply(Matrix44SSE output, Matrix44SSE mat1, Matrix44SSE mat0)
{
	output[0] = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	output[1] = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	output[2] = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	output[3] = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
}

LUX_INLINE void Matrix44SSE_multiply1(  Matrix44SSE mat1, Matrix44SSE mat0)
{
	__m128 m1 = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	__m128 m2 = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	__m128 m3 = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	__m128 m4 = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	mat1[0] = m1;
	mat1[1] = m2;
	mat1[2] = m3;
	mat1[3] = m4;
}
LUX_INLINE void Matrix44SSE_multiply2( Matrix44SSE mat1, Matrix44SSE mat0)
{

	mat0[0] = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	mat0[1] = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	mat0[2] = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	mat0[3] = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(3,3,3,3)), mat1[3]));

}


LUX_INLINE void Matrix44SSE_affineInvert(Matrix44SSE outmat, Matrix44SSE mat)
{
	outmat[0] = mat[0];
	outmat[1] = mat[1];
	outmat[2] = mat[2];
	outmat[3] = mat[3];

	_MM_TRANSPOSE4_PS(outmat[0], outmat[1], outmat[2], outmat[3]);

	outmat[3] = _mm_add_ps(
		_mm_add_ps(
		_mm_mul_ps(
		_mm_shuffle_ps(mat[3], mat[3], _MM_SHUFFLE(0,0,0,0)), outmat[0]), 
		_mm_mul_ps(
		_mm_shuffle_ps(mat[3], mat[3], _MM_SHUFFLE(1,1,1,1)), outmat[1])), 
		_mm_mul_ps(
		_mm_shuffle_ps(mat[3], mat[3], _MM_SHUFFLE(2,2,2,2)), outmat[2]));

	Vector4SSE_negated(&outmat[3]);

	outmat[0].m128_f32[3] = 0.0f;
	outmat[1].m128_f32[3] = 0.0f;
	outmat[2].m128_f32[3] = 0.0f;
	outmat[3].m128_f32[3] = 1.0f;

}

LUX_INLINE void Matrix44SSE_invert(Matrix44SSE outmat, Matrix44SSE mat)
{
	float* src = (float*)mat;
	float* dst = (float*)outmat;

	__m128 minor0, minor1, minor2, minor3;
	__m128 row0, row1, row2, row3;
	__m128 det, tmp1;

#ifdef _DEBUG
	row1 = row3 = tmp1 = mat[0];
#endif

	tmp1 = _mm_loadh_pi(_mm_loadl_pi(tmp1, (__m64*)(src)), (__m64*)(src+ 4));
	row1 = _mm_loadh_pi(_mm_loadl_pi(row1, (__m64*)(src+8)), (__m64*)(src+12));

	row0 = _mm_shuffle_ps(tmp1, row1, 0x88);
	row1 = _mm_shuffle_ps(row1, tmp1, 0xDD);

	tmp1 = _mm_loadh_pi(_mm_loadl_pi(tmp1, (__m64*)(src+ 2)), (__m64*)(src+ 6));
	row3 = _mm_loadh_pi(_mm_loadl_pi(row3, (__m64*)(src+10)), (__m64*)(src+14));

	row2 = _mm_shuffle_ps(tmp1, row3, 0x88);
	row3 = _mm_shuffle_ps(row3, tmp1, 0xDD);

	tmp1 = _mm_mul_ps(row2, row3);
	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0xB1);

	minor0 = _mm_mul_ps(row1, tmp1);
	minor1 = _mm_mul_ps(row0, tmp1);

	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0x4E);

	minor0 = _mm_sub_ps(_mm_mul_ps(row1, tmp1), minor0);
	minor1 = _mm_sub_ps(_mm_mul_ps(row0, tmp1), minor1);
	minor1 = _mm_shuffle_ps(minor1, minor1, 0x4E);

	tmp1 = _mm_mul_ps(row1, row2);
	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0xB1);

	minor0 = _mm_add_ps(_mm_mul_ps(row3, tmp1), minor0);
	minor3 = _mm_mul_ps(row0, tmp1);

	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0x4E);

	minor0 = _mm_sub_ps(minor0, _mm_mul_ps(row3, tmp1));
	minor3 = _mm_sub_ps(_mm_mul_ps(row0, tmp1), minor3);
	minor3 = _mm_shuffle_ps(minor3, minor3, 0x4E);

	tmp1 = _mm_mul_ps(_mm_shuffle_ps(row1, row1, 0x4E), row3);
	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0xB1);
	row2 = _mm_shuffle_ps(row2, row2, 0x4E);

	minor0 = _mm_add_ps(_mm_mul_ps(row2, tmp1), minor0);
	minor2 = _mm_mul_ps(row0, tmp1);

	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0x4E);

	minor0 = _mm_sub_ps(minor0, _mm_mul_ps(row2, tmp1));
	minor2 = _mm_sub_ps(_mm_mul_ps(row0, tmp1), minor2);
	minor2 = _mm_shuffle_ps(minor2, minor2, 0x4E);

	tmp1 = _mm_mul_ps(row0, row1);
	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0xB1);

	minor2 = _mm_add_ps(_mm_mul_ps(row3, tmp1), minor2);
	minor3 = _mm_sub_ps(_mm_mul_ps(row2, tmp1), minor3);

	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0x4E);

	minor2 = _mm_sub_ps(_mm_mul_ps(row3, tmp1), minor2);
	minor3 = _mm_sub_ps(minor3, _mm_mul_ps(row2, tmp1));

	tmp1 = _mm_mul_ps(row0, row3);
	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0xB1);

	minor1 = _mm_sub_ps(minor1, _mm_mul_ps(row2, tmp1));
	minor2 = _mm_add_ps(_mm_mul_ps(row1, tmp1), minor2);

	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0x4E);

	minor1 = _mm_add_ps(_mm_mul_ps(row2, tmp1), minor1);
	minor2 = _mm_sub_ps(minor2, _mm_mul_ps(row1, tmp1));

	tmp1 = _mm_mul_ps(row0, row2);
	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0xB1);

	minor1 = _mm_add_ps(_mm_mul_ps(row3, tmp1), minor1);
	minor3 = _mm_sub_ps(minor3, _mm_mul_ps(row1, tmp1));

	tmp1 = _mm_shuffle_ps(tmp1, tmp1, 0x4E);

	minor1 = _mm_sub_ps(minor1, _mm_mul_ps(row3, tmp1));
	minor3 = _mm_add_ps(_mm_mul_ps(row1, tmp1), minor3);

	det = _mm_mul_ps(row0, minor0);
	det = _mm_add_ps(_mm_shuffle_ps(det, det, 0x4E), det);
	det = _mm_add_ss(_mm_shuffle_ps(det, det, 0xB1), det);
	tmp1 = _mm_rcp_ss(det);

	det = _mm_sub_ss(_mm_add_ss(tmp1, tmp1), _mm_mul_ss(det, _mm_mul_ss(tmp1, tmp1)));
	det = _mm_shuffle_ps(det, det, 0x00);

	minor0 = _mm_mul_ps(det, minor0);
	_mm_storel_pi((__m64*)(dst), minor0);
	_mm_storeh_pi((__m64*)(dst+2), minor0);

	minor1 = _mm_mul_ps(det, minor1);
	_mm_storel_pi((__m64*)(dst+4), minor1);
	_mm_storeh_pi((__m64*)(dst+6), minor1);

	minor2 = _mm_mul_ps(det, minor2);
	_mm_storel_pi((__m64*)(dst+ 8), minor2);
	_mm_storeh_pi((__m64*)(dst+10), minor2);

	minor3 = _mm_mul_ps(det, minor3);
	_mm_storel_pi((__m64*)(dst+12), minor3);
	_mm_storeh_pi((__m64*)(dst+14), minor3);
}

//////////////////////////////////////////////////////////////////////////
// Matrix34SSE
LUX_INLINE void Matrix34SSE_identity(Matrix34SSE outmat)
{
	outmat[0] = _mm_set_ps(0.0f,0.0f,0.0f,1.0f);
	outmat[1] = _mm_set_ps(0.0f,0.0f,1.0f,0.0f);
	outmat[2] = _mm_set_ps(0.0f,1.0f,0.0f,0.0f);
}

LUX_INLINE void Matrix34SSE_copy(Matrix34SSE outmat, Matrix34SSE src)
{
	outmat[0] = src[0];
	outmat[1] = src[1];
	outmat[2] = src[2];
}

LUX_INLINE void Matrix34SSE_tmultiply44(Matrix34SSE outmat, Matrix44SSE mat1,  Matrix44SSE mat0)
{
	__m128 row0 = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[0], mat0[0], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	__m128 row1 = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[1], mat0[1], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	__m128 row2 = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[2], mat0[2], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
	__m128 row3 = _mm_add_ps(_mm_add_ps(_mm_add_ps(_mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(0,0,0,0)), mat1[0]), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(1,1,1,1)), mat1[1])), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(2,2,2,2)), mat1[2])), _mm_mul_ps(_mm_shuffle_ps(mat0[3], mat0[3], _MM_SHUFFLE(3,3,3,3)), mat1[3]));
            
	__m128 tmp3, tmp2, tmp1, tmp0;                          
	
	tmp0   = _mm_shuffle_ps((row0), (row1), 0x44);          
	tmp2   = _mm_shuffle_ps((row0), (row1), 0xEE);          
	tmp1   = _mm_shuffle_ps((row2), (row3), 0x44);          
	tmp3   = _mm_shuffle_ps((row2), (row3), 0xEE);          
	
	outmat[0] = _mm_shuffle_ps(tmp0, tmp1, 0x88);              
	outmat[1] = _mm_shuffle_ps(tmp0, tmp1, 0xDD);              
	outmat[2] = _mm_shuffle_ps(tmp2, tmp3, 0x88);              
}


//////////////////////////////////////////////////////////////////////////
//
/*
LUX_INLINE void QuatToMatrixSSE(const Quat q,Matrix44SSE a_mat)
{
	// Quat to mat conversion, calculate coefficients
	float x2 = q[0]+ q[0];
	float y2 = q[1]+ q[1];
	float z2 = q[2]+ q[2];
	float xx = q[0]* x2;
	float xy = q[0]* y2;
	float xz = q[0]* z2;
	float yy = q[1]* y2;
	float yz = q[1]* z2;
	float zz = q[2]* z2;
	float wx = q[3]* x2;
	float wy = q[3]* y2;
	float wz = q[3]* z2;

	// fill mat
	a_mat[0] = 1.0f- (yy+ zz);
	a_mat[4] = xy- wz;
	a_mat[8] = xz+ wy;

	a_mat[1] = xy+ wz;
	a_mat[5] = 1.0f- (xx+ zz);
	a_mat[9] = yz- wx;

	a_mat[2] = xz- wy;
	a_mat[6] = yz+ wx;
	a_mat[10]= 1.0f- (xx+ yy);
	
}
*/
#endif

#ifdef __cplusplus
}
#endif

#endif
