// Copyright (C) 2004-2009 Christoph Kubisch & Eike Decker
// This file is part of the "Luxinia Engine".
// For conditions of distribution and use, see copyright notice in luxiniapublicsdk.txt

#include <luxmath/bounding.h>
#include <luxmath/vector3.h>

// Bounding Volumes
// ----------------

void BoundingBox_complete(BoundingBox_t *bbox, booln fromMinMax){
	// we have min max and create center + length
	// or opposite
	if (fromMinMax){
		Vector3Sub(bbox->center,bbox->max, bbox->min);
		bbox->center[0] /= 2;
		bbox->center[1] /= 2;
		bbox->center[2] /= 2;
		Vector3Add(bbox->center,bbox->center, bbox->min);

		Vector3Sub(bbox->length,bbox->max,bbox->min);
	}
	else{
		Vector3ScaledAdd(bbox->max,bbox->center,0.5,bbox->length);
		Vector3ScaledAdd(bbox->min,bbox->center,-0.5,bbox->length);
	}

}

BoundingBox_t *BoundingBox_copy(BoundingBox_t *out, BoundingBox_t *in)
{
	Vector3Copy(out->max, in->max);
	Vector3Copy(out->min, in->min);
	Vector3Copy(out->center, in->center);
	Vector3Copy(out->length,in->length);

	return out;
}

BoundingSphere_t *BoundingSphere_copy(BoundingSphere_t *out, const BoundingSphere_t *in)
{
	Vector3Copy(out->center, in->center);
	out->radius = in->radius;

	return out;
}
BoundingBox_t * BoundingBox_merge(BoundingBox_t *out, BoundingBox_t *a, BoundingBox_t *b)
{
	out->min[0] = MIN(a->min[0], b->min[0]);
	out->min[1] = MIN(a->min[1], b->min[1]);
	out->min[2] = MIN(a->min[2], b->min[2]);

	out->max[0] = MAX(a->max[0], b->max[0]);
	out->max[1] = MAX(a->max[1], b->max[1]);
	out->max[2] = MAX(a->max[2], b->max[2]);

	return out;
}
booln BoundingBox_mergeChange(BoundingBox_t *out, BoundingBox_t *a, BoundingBox_t *b)
{
	int change = FALSE;
	float temp;
	temp = MIN(a->min[0], b->min[0]);
	change |= (a->min[0] != temp);
	out->min[0] = temp;
	temp = MIN(a->min[1], b->min[1]);
	change |= (a->min[1] != temp);
	out->min[1] = temp;
	temp = MIN(a->min[2], b->min[2]);
	change |= (a->min[2] != temp);
	out->min[2] = temp;

	temp = MAX(a->max[0], b->max[0]);
	change |= (a->max[0] != temp);
	out->max[0] = temp;
	temp = MAX(a->max[1], b->max[1]);
	change |= (a->max[1] != temp);
	out->max[1] = temp;
	temp = MAX(a->max[2], b->max[2]);
	change |= (a->max[2] != temp);
	out->max[2] = temp;

	return change;
}
booln BoundingBox_intersect(const BoundingBox_t *a,const BoundingBox_t *b)
{
	return !( (a)->min[0]>(b)->max[0] || (a)->min[1]>(b)->max[1] ||
		(a)->min[2]>(b)->max[2] || (a)->max[0]<(b)->min[0] ||
		(a)->max[1]<(b)->min[1] || (a)->max[2]<(b)->min[2]	);
}

booln BoundingBox_checkPoint(const BoundingBox_t *out, const Vector3 pt){
	if (pt[0] < out->min[0] || pt[0] > out->max[0] ||
		pt[1] < out->min[1] || pt[1] > out->max[1] ||
		pt[2] < out->min[2] || pt[2] > out->max[2])
		return FALSE;
	return TRUE;
}

booln BoundingSphere_merge(BoundingSphere_t *out, const BoundingSphere_t *a, const BoundingSphere_t *b)
{
	Vector3 vec;
	float rad;

	// if one sphere has a radius of 0 return the other sphere
	if(a->radius == 0){
		BoundingSphere_copy(out, b);
		return TRUE;
	}
	if(b->radius == 0){
		BoundingSphere_copy(out, a);
		return FALSE;
	}

	// get vector from a->center to b->center
	Vector3Sub(vec,b->center, a->center);

	// get distance of the two centers and normalize the vector
	rad = Vector3NormalizedA(vec);

	// if one sphere is in the other return the larger sphere
	if(rad + MIN(a->radius, b->radius) <= MAX(a->radius, b->radius))
	{
		if(a->radius > b->radius){
			BoundingSphere_copy(out, a);
			return FALSE;
		}
		else{
			BoundingSphere_copy(out, b);
			return TRUE;
		}
	}

	// calculate new radius
	rad += a->radius + b->radius;
	rad *= .5;

	// calculate new center
	Vector3Scale(vec,vec, rad-a->radius);
	Vector3Add(vec,vec, a->center);

	// copy the new sphere
	Vector3Copy(out->center, vec);
	out->radius = rad;

	return TRUE;
}

void BoundingBox_toSphereV(const Vector3 min,const Vector3 max, Vector3 center, float *radius)
{
	// center = (max-min)/2 + min
	Vector3Sub(center,max, min);
	center[0] *= 0.5f;
	center[1] *= 0.5f;
	center[2] *= 0.5f;

	// radius = len((max-min)/2)
	*radius = Vector3Length(center);

	Vector3Add( center,center, min);
}

BoundingSphere_t *BoundingBox_toSphere(BoundingBox_t *bbox, BoundingSphere_t *sphere)
{
	BoundingBox_toSphereV(bbox->min, bbox->max, sphere->center, &sphere->radius);
	sphere->radiusSqr = sphere->radius * sphere->radius;
	return sphere;
}



BoundingCapsule_t *BoundingBox_toCapsule(BoundingBox_t *bbox,BoundingCapsule_t *capsule)
{
	int		minaxis = 0;
	float	t;
	float	mindim = bbox->length[0];

	if (bbox->length[1] < mindim){
		mindim = bbox->length[1];
		minaxis = 1;
	}
	if (bbox->length[2] < mindim){
		mindim = bbox->length[2];
		minaxis = 2;
	}

	// pick the greater dimension of the other 2 axis as radius
	mindim = bbox->length[(minaxis+1)%3];
	mindim*=0.5f;
	t = bbox->length[(minaxis+2)%3];
	t*=0.5f;
	capsule->radius = (float)sqrt(mindim*mindim + t*t);
	Vector3Clear(capsule->toEnd);
	// length of line segment is "length - 2 * radius"
	capsule->toEnd[minaxis] = bbox->length[minaxis];

	//capsule->radius*=0.5f;
	capsule->radiusSqr = capsule->radius*capsule->radius;

	// create start point
	Vector3Copy(capsule->origin,bbox->center);
	capsule->origin[minaxis] = capsule->origin[minaxis]- bbox->length[minaxis]*0.5f;

	return capsule;
}



void BoundingBox_transformBoxCorners(const BoundingBox_t *in, const Matrix44 trans, Vector3 box[8])
{
	int i;
	// create box corners
	Vector3Copy(box[0],in->min);
	Vector3Copy(box[1],in->min);
	box[1][0] = in->max[0];
	Vector3Copy(box[2],box[0]);
	Vector3Copy(box[3],box[1]);
	box[2][1] = box[3][1] = in->max[1];
	Vector3Copy(box[4],in->max);
	Vector3Copy(box[5],in->max);
	box[5][0] = in->min[0];
	Vector3Copy(box[6],box[4]);
	Vector3Copy(box[7],box[5]);
	box[6][1] = box[7][1] = in->min[1];

	for (i = 0; i < 8 ; i++)
		Vector3Transform1(box[i],trans);
}

void BoundingBox_transformV(Vector3 outmins, Vector3 outmaxs,Vector3 mins, Vector3 maxs, Matrix44 trans)
{
	int i;
	Vector3 test;
	Vector3	box[8];
	// create box corners
	Vector3Copy(box[0],mins);
	Vector3Copy(box[1],mins);
	box[1][0] = maxs[0];
	Vector3Copy(box[2],box[0]);
	Vector3Copy(box[3],box[1]);
	box[2][1] = box[3][1] = maxs[1];
	Vector3Copy(box[4],maxs);
	Vector3Copy(box[5],maxs);
	box[5][0] = mins[0];
	Vector3Copy(box[6],box[4]);
	Vector3Copy(box[7],box[5]);
	box[6][1] = box[7][1] = mins[1];
	// transform box corners
	// and find new mins,maxs

	Vector3Transform(test,box[0],trans);
	Vector3Set(outmins,test[0],test[1],test[2]);
	Vector3Set(outmaxs,test[0],test[1],test[2]);
	for (i = 1; i < 8 ; i++){
		Vector3Transform(test,box[i],trans);

		outmins[0] = MIN(outmins[0],test[0]);
		outmins[1] = MIN(outmins[1],test[1]);
		outmins[2] = MIN(outmins[2],test[2]);

		outmaxs[0] = MAX(outmaxs[0], test[0]);
		outmaxs[1] = MAX(outmaxs[1], test[1]);
		outmaxs[2] = MAX(outmaxs[2], test[2]);
	}
}

BoundingBox_t *BoundingBox_transform(BoundingBox_t *out,BoundingBox_t *in, Matrix44 trans)
{
	BoundingBox_transformV(out->min,out->max,in->min, in->max, trans);
	return out;
}

void BoundingBox_fromCorners(BoundingBox_t *bbox,const Vector3 vecs[8])
{
	const float *test;
	int i;

	test = vecs[0];
	Vector3Set(bbox->min,test[0],test[1],test[2]);
	Vector3Set(bbox->max,test[0],test[1],test[2]);

	for (i = 1; i < 8 ; i++){
		test = vecs[i];
		bbox->min[0] = MIN(bbox->min[0],test[0]);
		bbox->min[1] = MIN(bbox->min[1],test[1]);
		bbox->min[2] = MIN(bbox->min[2],test[2]);

		bbox->max[0] = MAX(bbox->max[0], test[0]);
		bbox->max[1] = MAX(bbox->max[1], test[1]);
		bbox->max[2] = MAX(bbox->max[2], test[2]);
	}
}

void BoundingVectors_fromCamera(Vector3 box[8],const Matrix44 mat,const float fov, const float frontplane, const float backplane, const float aspect)
{
	int i;
	float invaspect = 1.0f/aspect;
	if (fov < 0.0f){
		Vector3Set(box[0],fov*aspect*0.5f,frontplane,fov*invaspect*0.5f);
		Vector3Set(box[1],-box[0][0],box[0][1],box[0][2]);
		Vector3Set(box[2],-box[0][0],box[0][1],-box[0][2]);
		Vector3Set(box[3],box[0][0],box[0][1],-box[0][2]);

		Vector3Set(box[4],box[0][0],backplane,box[0][2]);
		Vector3Set(box[5],-box[4][0],box[4][1],box[4][2]);
		Vector3Set(box[6],-box[4][0],box[4][1],-box[4][2]);
		Vector3Set(box[7],box[4][0],box[4][1],-box[4][2]);
	}
	else{
		float tfov = (float)tan(DEG2RAD(fov*0.5f));
		float tifov = (float)tan(DEG2RAD(fov*invaspect*0.5f));

		Vector3Set(box[0],tfov*frontplane,frontplane,tifov*frontplane);
		Vector3Set(box[1],-box[0][0],box[0][1],box[0][2]);
		Vector3Set(box[2],-box[0][0],box[0][1],-box[0][2]);
		Vector3Set(box[3],box[0][0],box[0][1],-box[0][2]);

		Vector3Set(box[4],tfov*backplane,backplane,tifov*backplane);
		Vector3Set(box[5],-box[4][0],box[4][1],box[4][2]);
		Vector3Set(box[6],-box[4][0],box[4][1],-box[4][2]);
		Vector3Set(box[7],box[4][0],box[4][1],-box[4][2]);
	}


	for (i=0; i < 8; i++){
		Vector3Transform1(box[i],mat);
	}
}



booln BoundingCone_checkSphereV(BoundingCone_t *cone,Vector3 center, float radius, float radiusSqr)
{
	Vector3 d;

	float flt,e;

	flt = radius * cone->sinDiv;

	Vector3ScaledAdd(d,cone->top,-flt,cone->axis);
	Vector3Sub(d,center,d);

	flt = Vector3Dot(d,d);
	e = Vector3Dot(cone->axis,d);
	if ( e > 0 && e*e >= flt*cone->cosSqr )
	{
		Vector3Sub(d,center,cone->top);
		flt = Vector3Dot(d,d);
		e = -Vector3Dot(cone->axis,d);
		if ( e > 0 && e*e >= flt*cone->sinSqr)
			return (flt <= radiusSqr);
		else
			return TRUE;
	}
	return FALSE;
}

void BoundingSphereCone_fromCamera(BoundingSphere_t *sphere, BoundingCone_t *cone,float frontplane, float backplane, Vector3 pos, Vector3 dir, float fov){
	Vector3 p,q,diff;
	float fFov = DEG2RAD(fov);
	float fViewLen = backplane - frontplane;
	// use some trig to find the height of the frustum at the far plane
	float fHeight = fViewLen * (float)tan(fFov * 0.5f);
	// with an aspect ratio of 1, the width will be the same
	float fWidth = fHeight;
	float fDepth  = fHeight / (float)tan(fFov * 0.5f);

	// calculate the corner of the screen
	float fCorner = fastsqrt(fWidth * fWidth + fHeight * fHeight);

	// now calculate the new fov
	fFov = (float)atan(fCorner / fDepth);

	// halfway point between near/far planes starting at the origin and extending along the z axis
	Vector3Set (p,0.0f, 0.0f, -(frontplane + fViewLen * 0.5f));

	// the calculate far corner of the frustum
	Vector3Set (q,fWidth, fHeight, fViewLen);

	// the vector between P and Q
	Vector3Sub(diff,p,q);

	// the radius becomes the length of this vector
	if (fov < 0.0f)
		sphere->radius = (-fov)*2.0f;
	else
		sphere->radius = Vector3Length(diff);
	sphere->radiusSqr = sphere->radius * sphere->radius;

	// calculate the center of the sphere
	fViewLen = fViewLen * 0.5f + frontplane;
	Vector3ScaledAdd(sphere->center,pos,fViewLen,dir);

	// apply to the cone
	Vector3Copy(cone->axis,dir);
	Vector3Copy(cone->top,pos);
	cone->cosSqr = fastcos(fFov);
	cone->cosSqr *= cone->cosSqr;
	cone->sinSqr = fastsin(fFov);
	cone->sinDiv = 1/cone->sinSqr;
	cone->sinSqr *= cone->sinSqr;
}

void BoundingSphere_fromFrustumCorners(BoundingSphere_t *sphere,Vector3 box[FRUSTUM_CORNERS])
{
	Vector3 ctr = {0.0f,0.0f,0.0f};
	float rad = 0.0f;
	int i;
	for (i = 0; i < FRUSTUM_CORNERS; i++){
		Vector3Add(ctr,ctr,box[i]);
	}
	Vector3Scale(ctr,ctr,1.0f/((float)FRUSTUM_CORNERS));

	for (i = 0; i < FRUSTUM_CORNERS; i++){
		Vector3 temp;
		float cur;
		Vector3Sub(temp,ctr,box[i]);
		cur = Vector3Dot(temp,temp);
		rad = MAX(rad,cur);
	}

	Vector3Copy(sphere->center,ctr);
	sphere->radiusSqr = rad;
	sphere->radius = fastsqrt(rad);
}

void BoundingCone_fromFrustumCorners(BoundingCone_t *cone, Vector3 box[FRUSTUM_CORNERS])
{
	Vector3 near;
	Vector3 far;
	Vector3 temp;
	Vector3 ctr;
	float radn = 0.0f;
	float radf = 0.0f;
	float rad = 0.0f;
	int i;

	Vector3Add(near,box[FRUSTUM_C_NTR],box[FRUSTUM_C_NTL]);
	Vector3Add(near,near,box[FRUSTUM_C_NBL]);
	Vector3Add(near,near,box[FRUSTUM_C_NBR]);
	Vector3Scale(near,near,0.25f);

	Vector3Add(far,box[FRUSTUM_C_FTR],box[FRUSTUM_C_FTL]);
	Vector3Add(far,far,box[FRUSTUM_C_FBL]);
	Vector3Add(far,far,box[FRUSTUM_C_FBR]);
	Vector3Scale(far,far,0.25f);

	for (i = 0; i < 4; i++){
		float distsq;
		Vector3Sub(temp,near,box[FRUSTUM_C_NTR+i]);
		distsq = Vector3Dot(temp,temp);
		radn = MAX(radn,distsq);

		Vector3Sub(temp,far,box[FRUSTUM_C_FTR+i]);
		distsq = Vector3Dot(temp,temp);
		radf = MAX(radf,distsq);
	}

	Vector3Sub(temp,near,far);
	radn = fastsqrt(radn);
	radf = fastsqrt(radf);
	rad = radn/radf;

	if (rad < 0.95f){
		Vector3Scale(ctr,temp,1.0f/(1.0f-rad));
		rad = Vector3LengthA(ctr);
		Vector3Add(cone->top,ctr,far);
		Vector3Scale(cone->axis,ctr,-1.0f/rad);
		rad = (float)atan(radf/rad)*2.0f;
		radf = fastcos(rad);
		cone->cosSqr = radf*radf;
		radf = fastsin(rad);
		cone->sinSqr = radf*radf;
		cone->sinDiv = 1.0f/radf;
	}
	else{
		cone->cosSqr = 1.0f;
		cone->sinDiv = 1.0f;
		cone->sinSqr = 0.0f;
	}
}


